import copy
import os
from pathlib import Path

from typing import Any, Dict, Optional, Union

from runhouse.globals import rns_client
from runhouse.logger import get_logger
from runhouse.resources.hardware.cluster import Cluster
from runhouse.resources.secrets.provider_secrets.provider_secret import ProviderSecret

logger = get_logger(__name__)


class SSHSecret(ProviderSecret):
    """
    .. note::
            To create a SSHSecret, please use the factory method :func:`provider_secret` with ``provider="ssh"``.
    """

    _DEFAULT_CREDENTIALS_PATH = "~/.ssh"
    _PROVIDER = "ssh"
    _DEFAULT_KEY = "id_rsa"

    def __init__(
        self,
        name: Optional[str] = None,
        provider: Optional[str] = None,
        values: Dict = {},
        path: str = None,
        key: str = None,
        dryrun: bool = True,
        **kwargs,
    ):
        self.key = (
            key or os.path.basename(path) if path else (name or self._DEFAULT_KEY)
        )
        super().__init__(
            name=name, provider=provider, values=values, path=path, dryrun=dryrun
        )
        if self.path == self._DEFAULT_CREDENTIALS_PATH:
            self.path = str(Path(self._DEFAULT_CREDENTIALS_PATH) / self.key)

    @staticmethod
    def from_config(config: dict, dryrun: bool = False, _resolve_children: bool = True):
        # try block if for the case we are trying to load a shared secret.
        return SSHSecret(**config, dryrun=dryrun)

    def save(
        self,
        name: str = None,
        save_values: bool = True,
        headers: Optional[Dict] = None,
        folder: str = None,
    ):
        if name:
            self.name = name
        elif not self.name:
            self.name = f"ssh-{self.key}"
            # Append username if available
            if rns_client.username:
                self.name = f"/{rns_client.username}/{self.name}"

        if self.path:
            try:
                rel_path = "~" / Path(self.path).relative_to(Path.home())
                self.path = str(rel_path)
            except (ValueError, RuntimeError):
                pass

        return super().save(
            save_values=save_values,
            headers=headers or rns_client.request_headers(),
            folder=folder,
        )

    def _write_to_file(
        self,
        path: str,
        values: Dict = None,
        overwrite: bool = False,
        write_config: bool = True,
    ):
        priv_key_path = path

        priv_key_path = Path(os.path.expanduser(priv_key_path))
        pub_key_path = Path(f"{os.path.expanduser(priv_key_path)}.pub")

        values = values or self.values

        if not overwrite and priv_key_path.exists() and pub_key_path.exists():
            if values == self._from_path(path=path):
                logger.info(f"Secrets already exist in {path}. Skipping.")
                self.path = path
                return self
            logger.warning(
                f"SSH Secrets for {self.name or self.key} already exist in {path}. "
                "Automatically overriding SSH keys is not supported by Runhouse. "
                "Please manually edit these files."
            )
            self.path = path
            return self

        priv_key_path.parent.mkdir(parents=True, exist_ok=True)
        private_key = values.get("private_key")
        if overwrite or (private_key is not None and not priv_key_path.exists()):
            priv_key_path.write_text(private_key)
            priv_key_path.chmod(0o600)
        public_key = values.get("public_key")
        if overwrite or (public_key is not None and not pub_key_path.exists()):
            pub_key_path.write_text(public_key)
            pub_key_path.chmod(0o600)

        new_secret = copy.deepcopy(self)
        new_secret._values = None
        new_secret.path = path
        new_secret.name = f"ssh-{os.path.basename(path)}"

        if write_config:
            try:
                new_secret._add_to_rh_config(val=path)
            except TypeError:
                pass

        return new_secret

    def _from_path(self, path: str):
        if path == self._DEFAULT_CREDENTIALS_PATH or path == os.path.expanduser(
            self._DEFAULT_CREDENTIALS_PATH
        ):
            path = f"{path}/{self.key}"

        return self.extract_secrets_from_path(path)

    @staticmethod
    def extract_secrets_from_path(path: str) -> Dict:
        pub_key_path = os.path.expanduser(f"{path}.pub")
        priv_key_path = os.path.expanduser(path)

        if not (os.path.exists(pub_key_path) and os.path.exists(priv_key_path)):
            return {}

        pub_key = Path(pub_key_path).read_text()
        priv_key = Path(priv_key_path).read_text()

        return {"public_key": pub_key, "private_key": priv_key}

    def _file_to(
        self,
        key: str,
        system: Union[str, Cluster],
        path: Union[str, Path] = None,
        values: Any = None,
    ):
        if self.path:
            remote_priv_file = self.path
            # pub_key_path = f"{path}.pub"
            system.call(key, "_write_to_file", path=remote_priv_file, values=values)
            system.run_bash_over_ssh([f"chmod 600 {path}"])
        else:
            system.call(key, "_write_to_file", path=path, values=values)
            remote_priv_file = path
        return remote_priv_file



from typing import Dict, Optional, Union

from runhouse.resources.hardware.cluster import Cluster
from runhouse.resources.secrets.provider_secrets.provider_secret import ProviderSecret


class ApiKeySecret(ProviderSecret):
    """Secret class for providers consisting of a single API key, generally stored as an environment variable.

    .. note::
            To create an ApiKeySecret, please use the factory method :func:`provider_secret`
            and passing in the corresponding provider.
    """

    def write(
        self,
        file: bool = False,
        env: bool = False,
        path: str = None,
        env_vars: Dict = None,
        overwrite: bool = False,
        write_config: bool = True,
    ):
        if not file or path:
            env = True
        super().write(
            file=file,
            env=env,
            path=path,
            env_vars=env_vars,
            overwrite=overwrite,
            write_config=write_config,
        )

    def to(
        self,
        system: Union[str, Cluster],
        path: str = None,
        process: Optional[str] = None,
        values: bool = True,
        name: Optional[str] = None,
    ):
        return super().to(
            system=system, path=path, process=process, values=values, name=name
        )



from runhouse.resources.secrets.provider_secrets.anthropic_secret import AnthropicSecret
from runhouse.resources.secrets.provider_secrets.aws_secret import AWSSecret
from runhouse.resources.secrets.provider_secrets.azure_secret import AzureSecret
from runhouse.resources.secrets.provider_secrets.cohere_secret import CohereSecret
from runhouse.resources.secrets.provider_secrets.docker_secret import (
    DockerRegistrySecret,
)
from runhouse.resources.secrets.provider_secrets.gcp_secret import GCPSecret
from runhouse.resources.secrets.provider_secrets.github_secret import GitHubSecret
from runhouse.resources.secrets.provider_secrets.huggingface_secret import (
    HuggingFaceSecret,
)
from runhouse.resources.secrets.provider_secrets.kubeconfig_secret import (
    KubeConfigSecret,
)
from runhouse.resources.secrets.provider_secrets.lambda_secret import LambdaSecret
from runhouse.resources.secrets.provider_secrets.langchain_secret import LangChainSecret
from runhouse.resources.secrets.provider_secrets.openai_secret import OpenAISecret
from runhouse.resources.secrets.provider_secrets.pinecone_secret import PineconeSecret
from runhouse.resources.secrets.provider_secrets.provider_secret import ProviderSecret
from runhouse.resources.secrets.provider_secrets.sky_secret import SkySecret
from runhouse.resources.secrets.provider_secrets.ssh_secret import SSHSecret
from runhouse.resources.secrets.provider_secrets.wandb_secret import WandBSecret


_str_to_provider_class = {
    # File and/or Env secrets
    "aws": AWSSecret,
    "azure": AzureSecret,
    "gcp": GCPSecret,
    "github": GitHubSecret,
    "huggingface": HuggingFaceSecret,
    "kubernetes": KubeConfigSecret,
    "lambda": LambdaSecret,
    "docker": DockerRegistrySecret,
    # SSH secrets
    "ssh": SSHSecret,
    "sky": SkySecret,
    # API key secrets
    "anthropic": AnthropicSecret,
    "cohere": CohereSecret,
    "langchain": LangChainSecret,
    "openai": OpenAISecret,
    "pinecone": PineconeSecret,
    "wandb": WandBSecret,
}


def _get_provider_class(provider_str):
    if provider_str not in _str_to_provider_class:
        return ProviderSecret
    return _str_to_provider_class[provider_str]



import pytest
import requests
import runhouse as rh

from tests.test_resources.test_modules.test_server_modules.assets.sample_fastapi_app import (
    app,
)


"""
Resources:
https://github.com/simonw/asgi-proxy-lib
https://github.com/valohai/asgiproxy
https://github.com/florimondmanca/awesome-asgi

TODO test with Den Auth and HTTPS enabled
"""


@pytest.mark.level("local")
async def test_asgi_server(cluster):
    fast_api_module = rh.asgi(app).to(cluster, name="fast_api_module")
    assert isinstance(fast_api_module, rh.Asgi)
    assert fast_api_module.summer(1, 2) == 3

    # Call an async method
    assert await fast_api_module.my_deeply_nested_async_endpoint("hello", 1, 2.0) == (
        "hello",
        1,
        2.0,
    )

    assert list(fast_api_module.my_streaming_endpoint()) == list(range(10))
    assert fast_api_module.my_endpoint_with_optional_body_params_and_header(
        a=1, b=2, c=3
    ) == (1, 2, 3)

    endpoint = fast_api_module.endpoint(external=False)
    res = requests.get(f"{endpoint}/summer/1", params={"b": 2})
    assert res.json() == 3
    res = requests.post(
        f"{endpoint}/my/deeply/hello/nested/endpoint/1",
        json={"arg3": 2.0},
    )
    assert res.json() == ["hello", 1, 2.0]
    res = requests.get(f"{endpoint}/my/streaming/endpoint")
    assert res.json() == list(range(10))
    res = requests.get(
        f"{endpoint}/my/endpoint/with/optional/body/params/and/header",
        json={"a": 1, "b": 2},
        headers={"c": 3},
    )
    assert res.json() == [1, 2, 3]

    # Test UI page pulls properly
    res = requests.get(f"{endpoint}/docs")
    assert res.status_code == 200


def test_python_script_start(cluster):
    mod = (
        tests.test_resources.test_modules.test_server_modules.assets.sample_fastapi_app
    )
    # Will start with python -m tests.test_resources.test_modules.test_server_modules.assets.sample_fastapi_app
    server_module = rh.server(mod, port=8000).to(
        cluster, env=["pytest", "requests"], name="fast_api_module"
    )
    assert isinstance(server_module, rh.WebServer)

    res = requests.get(
        "http://localhost:8001/fast_api_module/summer/1", params={"b": 2}
    )
    assert res.json() == 3
    res = requests.post(
        "http://localhost:8001/fast_api_module/my/deeply/hello/nested/endpoint/1",
        json={"arg3": 2.0},
    )
    assert res.json() == ["hello", 1, 2.0]
    res = requests.get("http://localhost:8001/fast_api_module/my/streaming/endpoint")
    assert res.json() == list(range(10))
    res = requests.get(
        "http://localhost:8001/fast_api_module/my/endpoint/with/optional/body/params/and/header",
        json={"a": 1, "b": 2},
        headers={"c": 3},
    )
    assert res.json() == [1, 2, 3]

    # Test UI page pulls properly
    res = requests.get("http://localhost:8000/fast_api_module/docs")
    assert res.status_code == 200

    fast_api_module.stop()

    # Test that the server is stopped
    with pytest.raises(requests.exceptions.ConnectionError):
        requests.get("http://localhost:8001/fast_api_module/summer/1", params={"b": 2})

    server_module.stop()


def test_custom_start_cmd(cluster):
    # Will start with python -m tests.test_resources.test_modules.test_server_modules.assets.sample_fastapi_app
    start_cmd = (
        "uvicorn --port 8001 --host 0.0.0.0 "
        "runhouse.tests.test_resources.test_modules.test_server_modules.assets.sample_fastapi_app:app"
    )
    server_module = rh.server(start_cmd, port=8000).to(
        cluster, env=["pytest", "requests"], name="fast_api_module"
    )
    assert isinstance(server_module, rh.WebServer)

    res = requests.get(
        "http://localhost:8001/fast_api_module/summer/1", params={"b": 2}
    )
    assert res.json() == 3
    res = requests.post(
        "http://localhost:8001/fast_api_module/my/deeply/hello/nested/endpoint/1",
        json={"arg3": 2.0},
    )
    assert res.json() == ["hello", 1, 2.0]
    res = requests.get("http://localhost:8001/fast_api_module/my/streaming/endpoint")
    assert res.json() == list(range(10))
    res = requests.get(
        "http://localhost:8001/fast_api_module/my/endpoint/with/optional/body/params/and/header",
        json={"a": 1, "b": 2},
        headers={"c": 3},
    )

    # Test UI page pulls properly
    res = requests.get("http://localhost:8000/fast_api_module/docs")
    assert res.status_code == 200

    fast_api_module.stop()
    with pytest.raises(requests.exceptions.ConnectionError):
        requests.get("http://localhost:8001/fast_api_module/summer/1", params={"b": 2})

    server_module.stop()



import pprint
import sys
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

from runhouse.globals import obj_store, rns_client
from runhouse.logger import get_logger

from runhouse.rns.top_level_rns_fns import (
    resolve_rns_path,
    save,
    split_rns_name_and_path,
)
from runhouse.rns.utils.api import (
    load_resp_content,
    read_resp_data,
    ResourceAccess,
    ResourceVisibility,
)
from runhouse.rns.utils.names import is_valid_resource_name

logger = get_logger(__name__)


class Resource:
    RESOURCE_TYPE = "resource"

    def __init__(
        self,
        name: Optional[str] = None,
        dryrun: bool = False,
        access_level: ResourceAccess = ResourceAccess.WRITE,
        visibility: ResourceVisibility = ResourceVisibility.PRIVATE,
        **kwargs,
    ):
        """
        Runhouse abstraction for objects that can be saved, shared, and reused.

        Args:
            name (Optional[str], optional): Name to assign the resource. (Default: None)
            dryrun (bool, optional): Whether to create the resource object, or load the object as a dryrun.
                (Default: ``False``)
            access_level (:obj:`ResourceAccess`, optional): Access level to provide for the resource.
                (Default: ``ResourceAccess.WRITE``)
            visibility (:obj:`ResourceVisibility`, optional): Type of visibility to provide for the resource.
                (Default: ``ResourceVisibility.PRIVATE``)
        """
        self._name, self._rns_folder = None, None
        if name is not None:
            if name.startswith("/builtins/"):
                name = name[len("/builtins/") :]
            if name[0] == "^" and name != "^":
                name = name[1:]
            # Validate that name complies with a simple regex
            if not is_valid_resource_name(name):
                raise ValueError(
                    f"Invalid name: {name} "
                    "Resource names are limited to alphanumerics, dashes, and underscores. Max 200 characters. "
                    "Slashes may be used to specify folders and must be included as the first character."
                )
            self._name, self._rns_folder = rns_client.split_rns_name_and_path(
                rns_client.resolve_rns_path(name)
            )

        self.dryrun = dryrun
        self.access_level = access_level
        self._visibility = visibility

    # TODO add a utility to allow a parameter to be specified as "default" and then use the default value

    def config(self, condensed=True):
        config = {
            "name": self.rns_address or self.name,
            "resource_type": self.RESOURCE_TYPE,
            "resource_subtype": self.__class__.__name__,
        }
        self.save_attrs_to_config(
            config,
            [
                "visibility",  # Handles Enum to string conversion
            ],
        )
        return config

    def _resource_string_for_subconfig(
        self, resource: Union[None, str, "Resource"], condensed=True
    ):
        """Returns a string representation of a sub-resource for use in a config."""
        if resource is None or isinstance(resource, str):
            return resource
        if isinstance(resource, Resource):
            if condensed and resource.rns_address:
                # We operate on the assumption that rns_address is only populated once a resource has been saved.
                # That way, if rns_address is not None, we have reasonable likelihood that the resource was saved and
                # we can just pass the address. The only exception here is if the resource is a built-in.
                if resource.rns_address.startswith("^"):
                    # Fork the resource if it's a built-in and consider it a new resource
                    resource._rns_folder = None
                    return resource.config(condensed)
                return resource.rns_address
            else:
                # If the resource doesn't have an rns_address, we consider it unsaved and put the whole config into
                # the parent config.
                return resource.config(condensed)
        raise ValueError(
            f"Resource {resource} is not a valid sub-resource for {self.__class__.__name__}"
        )

    @property
    def rns_address(self):
        """Full address of resource saved in Den. Has the format {username}/{resource_name}"""
        if (
            self.name is None or self._rns_folder is None
        ):  # Anonymous folders have no rns address
            return None

        return str(Path(self._rns_folder) / self.name)

    @property
    def name(self):
        """Resource name."""
        return self._name

    @name.setter
    def name(self, name):
        # Split the name and rns path if path is given (concat with current_folder if just stem is given)
        if name is None:
            self._name = None
        else:
            self._name, self._rns_folder = split_rns_name_and_path(
                resolve_rns_path(name)
            )

    @property
    def visibility(self):
        """Resource visibility."""
        return self._visibility

    @visibility.setter
    def visibility(self, visibility):
        self._visibility = visibility

    @rns_address.setter
    def rns_address(self, new_address):
        self.name = new_address  # Note, this saves the resource to the new address!

    def _save_sub_resources(self, folder: str = None):
        """Overload by child resources to save any resources they hold internally."""
        pass

    def pin(self):
        """Write the resource to the object store."""
        from runhouse.resources.hardware.utils import _current_cluster

        if _current_cluster():
            if obj_store.has_local_storage:
                obj_store.put_local(self._name, self)
            else:
                obj_store.put(self._name, self)
        else:
            raise ValueError("Cannot pin a resource outside of a cluster.")

    def refresh(self):
        """Update the resource in the object store."""
        from runhouse.resources.hardware.utils import _current_cluster

        if _current_cluster():
            return obj_store.get(self._name)
        else:
            return self

    def save(self, name: str = None, overwrite: bool = True, folder: str = None):
        """Register the resource, saving it to the Den config store. Uses the resource's
        ``self.config()`` to generate the dict to save."""

        # add this resource this run's downstream artifact registry if it's being saved as part of a run
        rns_client.add_downstream_resource(name or self.name)

        self._save_sub_resources(folder)
        if name:
            self.name = name

        # TODO handle self.access == 'read' instead of this weird overwrite argument
        save(self, overwrite=overwrite, folder=folder)

        return self

    def __str__(self):
        return pprint.pformat(self.config())

    @classmethod
    def _check_for_child_configs(cls, config: dict):
        """Overload by child resources to load any resources they hold internally."""
        return config

    @classmethod
    def from_name(
        cls,
        name: str,
        load_from_den: bool = True,
        dryrun: bool = False,
        _resolve_children: bool = True,
    ):
        """Load existing Resource via its name.

        Args:
            name (str): Name of the resource to load from name.
            load_from_den (bool, optional): Whether to try loading the module from Den. (Default: ``True``)
            dryrun (bool, optional): Whether to construct the object or load as dryrun. (Default: ``False``)
        """
        # TODO is this the right priority order?
        from runhouse.resources.hardware.utils import _current_cluster

        if _current_cluster() and obj_store.contains(name):
            return obj_store.get(name)

        config = rns_client.load_config(name=name, load_from_den=load_from_den)
        if not config:
            raise ValueError(f"Resource {name} not found.")

        if _resolve_children:
            config = cls._check_for_child_configs(config)

        # Add this resource's name to the resource artifact registry if part of a run
        rns_client.add_upstream_resource(name)

        # Uses child class's from_config
        return cls.from_config(
            config=config, dryrun=dryrun, _resolve_children=_resolve_children
        )

    @staticmethod
    def from_config(config: Dict, dryrun: bool = False, _resolve_children: bool = True):
        """Load or construct resource from config.

        Args:
            config (Dict): Resource config.
            dryrun (bool, optional): Whether to construct resource or load as dryrun (Default: ``False``)
        """
        resource_type = config.pop("resource_type", None)
        dryrun = config.pop("dryrun", False) or dryrun

        if resource_type == "resource":
            return Resource(**config, dryrun=dryrun)

        resource_class = getattr(
            sys.modules["runhouse"], resource_type.capitalize(), None
        )
        if not resource_class:
            raise TypeError(f"Could not find module associated with {resource_type}")

        if _resolve_children:
            config = resource_class._check_for_child_configs(config)

        loaded = resource_class.from_config(
            config=config,
            dryrun=dryrun,
            _resolve_children=_resolve_children,
        )
        if loaded.name:
            rns_client.add_upstream_resource(loaded.name)
        return loaded

    def unname(self):
        """Remove the name of the resource. This changes the resource name to anonymous and deletes any Den configs
        for the resource."""
        self.delete_configs()
        self._name = None

    def history(self, limit: int = None) -> List[Dict]:
        """Return the history of the resource, including specific config fields (e.g. folder path) and which runs
        have overwritten it.

        Args:
            limit (int, optional): If specified, return the last ``limit`` number of entries in the history.
                Otherwise, return the entire history. (Default: ``None``)
        """
        if not self.rns_address:
            raise ValueError("Resource must have a name in order to have a history")

        if self.rns_address[:2] == "~/":
            raise ValueError(
                "Resource must be saved to Den (not local) in order to have a history"
            )

        resource_uri = rns_client.resource_uri(self.rns_address)
        base_uri = f"{rns_client.api_server_url}/resource/history/{resource_uri}"
        uri = f"{base_uri}?limit={limit}" if limit else base_uri

        resp = rns_client.session.get(uri, headers=rns_client.request_headers())
        if resp.status_code != 200:
            logger.warning(
                f"Received [{resp.status_code}] from Den GET '{uri}': No resource history found: {load_resp_content(resp)}"
            )
            return []

        resource_history = read_resp_data(resp)
        return resource_history

    # TODO delete sub-resources
    def delete_configs(self):
        """Delete the resource's config from Den config store."""
        rns_client.delete_configs(resource=self)

    def save_attrs_to_config(self, config: Dict, attrs: List[str]):
        """Save the given attributes to the config"""
        for attr in attrs:
            val = self.__getattribute__(attr)
            if val or (val is False):
                # allow for saving `False` but not other falsey types
                if isinstance(val, Enum):
                    val = val.value
                config[attr] = val

    def is_local(self):
        return (
            hasattr(self, "install_target")
            and isinstance(self.install_target, str)
            and self.install_target.startswith("~")
            or hasattr(self, "system")
            and self.system == "file"
        )

    # TODO [DG] Implement proper sharing of subresources (with an overload of some kind)
    def share(
        self,
        users: Union[str, List[str]] = None,
        access_level: Union[ResourceAccess, str] = ResourceAccess.READ,
        visibility: Optional[Union[ResourceVisibility, str]] = None,
        notify_users: bool = True,
        headers: Dict = None,
    ) -> Tuple[Dict[str, ResourceAccess], Dict[str, ResourceAccess]]:
        """Grant access to the resource for a list of users (or a single user). By default, the user will
        receive an email notification of access (if they have a Runhouse account) or instructions on creating
        an account to access the resource. If ``visibility`` is set to ``public``, users will not be notified.

        .. note::
            You can only grant access to other users if you have write access to the resource.

        Args:
            users (Union[str, list], optional): Single user or list of user emails and / or runhouse account usernames.
                If none are provided and ``visibility`` is set to ``public``, resource will be made publicly
                available to all users. (Default: ``None``)
            access_level (:obj:`ResourceAccess`, optional): Access level to provide for the resource.
                (Default: ``read``).
            visibility (:obj:`ResourceVisibility`, optional): Type of visibility to provide for the shared
                resource. By default, the visibility is private. (Default: ``None``)
            notify_users (bool, optional): Whether to send an email notification to users who have been given access.
                Note: This is relevant for resources which are not ``shareable``. (Default: ``True``)
            headers (Dict, optional): Request headers to provide for the request to Den. Contains the user's auth token.
                Example: ``{"Authorization": f"Bearer {token}"}``

        Returns:
            Tuple(Dict, Dict, Set):

            `added_users`:
                Users who already have a Runhouse account and have been granted access to the resource.
            `new_users`:
                Users who do not have Runhouse accounts and received notifications via their emails.
            `valid_users`:
                Set of valid usernames and emails from ``users`` parameter.

        Example:
            >>> # Write access to the resource for these specific users.
            >>> # Visibility will be set to private (users can search for and view resource in Den dashboard)
            >>> my_resource.share(users=["username1", "user2@gmail.com"], access_level='write')

            >>> # Make resource public, with read access to the resource for all users
            >>> my_resource.share(visibility='public')
        """
        if self.name is None:
            raise ValueError("Resource must have a name in order to share")

        if users is None and visibility is None:
            raise ValueError(
                "Must specify `visibility` for the resource if no users are provided."
            )

        if hasattr(self, "system") and self.system in ["ssh", "sftp"]:
            logger.warning(
                "Sharing a resource located on a cluster is not recommended. For persistence, we suggest"
                "saving to a cloud storage system (ex: `s3` or `gs`). You can copy your cluster based "
                f"{self.RESOURCE_TYPE} to your desired storage provider using the `.to()` method. "
                f"For example: `{self.RESOURCE_TYPE}.to(system='rh-cpu')`"
            )

        if self.is_local():
            if self.RESOURCE_TYPE == "package":
                raise TypeError(
                    f"Unable to share a local {self.RESOURCE_TYPE}. Please make sure the {self.RESOURCE_TYPE} is "
                    f"located on a cluster. You can use the `.to()` method to do so. "
                    f"For example: `{self.name}.to(system='rh-cpu')`"
                )
            else:
                raise TypeError(
                    f"Unable to share a local {self.RESOURCE_TYPE}. Please make sure the {self.RESOURCE_TYPE} is "
                    f"located on a cluster or a remote system. You can use the `.to()` method to do so. "
                    f"For example: `{self.name}.to(system='s3')`"
                )

        if isinstance(access_level, str):
            access_level = ResourceAccess(access_level)

        if visibility is not None:
            # Update the resource in Den with this global visibility value
            self.visibility = visibility

            logger.debug(f"Updating resource with visibility: {self.visibility}")

        self.save()

        if isinstance(users, str):
            users = [users]

        added_users, new_users, valid_users = rns_client.grant_resource_access(
            rns_address=self.rns_address,
            user_emails=users,
            access_level=access_level,
            notify_users=notify_users,
            headers=headers,
        )
        return added_users, new_users, valid_users

    def revoke(
        self, users: Union[str, List[str]] = None, headers: Optional[Dict] = None
    ):
        """Revoke access to the resource.

        Args:
            users (Union[str, str], optional): List of user emails and / or runhouse account usernames
                (or a single user). If no users are specified will revoke access for all users. (Default: ``None``)
            headers (Optional[Dict]): Request headers to provide for the request to Den. Contains the user's auth token.
                Example: ``{"Authorization": f"Bearer {token}"}`` Will default to authorization headers of the local
                config. (Default: ``None``)
        """
        if isinstance(users, str):
            users = [users]

        request_uri = rns_client.resource_uri(self.rns_address)
        uri = f"{rns_client.api_server_url}/resource/{request_uri}/users/access"
        resp = rns_client.session.put(
            uri,
            json={"users": users, "access_level": ResourceAccess.DENIED},
            headers=headers or rns_client.request_headers(),
        )

        if resp.status_code != 200:
            raise Exception(
                f"Received [{resp.status_code}] from Den PUT '{uri}': Failed to revoke access for resource: {load_resp_content(resp)}"
            )



import copy
import shutil
import subprocess
import time
from pathlib import Path
from typing import Dict, List, Optional, Union

from runhouse.logger import get_logger

from .folder import Folder

MAX_POLLS = 120000
POLL_INTERVAL = 1
TIMEOUT_SECONDS = 3600

logger = get_logger(__name__)


class S3Folder(Folder):
    RESOURCE_TYPE = "folder"
    DEFAULT_FS = "s3"

    def __init__(self, dryrun: bool, **kwargs):
        try:
            import boto3
        except ImportError:
            raise ImportError(
                "`boto3` is required for S3 folders. You can install it with `pip install boto3`."
            )

        super().__init__(dryrun=dryrun, **kwargs)
        self.client = boto3.client("s3")
        self._urlpath = "s3://"

    @staticmethod
    def from_config(config: Dict, dryrun: bool = False, _resolve_children: bool = True):
        """Load config values into the object."""
        return S3Folder(**config, dryrun=dryrun)

    def _to_local(self, dest_path: str):
        """Copies folder to local."""
        from runhouse import Cluster

        if self._fs_str == "file":
            shutil.copytree(src=self.path, dst=dest_path)
        elif isinstance(self.system, Cluster):
            return self._cluster_to_local(cluster=self.system, dest_path=dest_path)
        else:
            self._s3_copy_to_local(dest_path)

        return self._destination_folder(dest_path=dest_path, dest_system="file")

    def _s3_copy_to_local(self, dest_path: str):
        """Copy S3 folder to local."""
        Path(dest_path).mkdir(parents=True, exist_ok=True)

        bucket_name = self._bucket_name
        key = self._key

        s3_objects = self.client.list_objects_v2(Bucket=bucket_name, Prefix=key)

        for obj in s3_objects.get("Contents", []):
            obj_key = obj["Key"]
            dest_file_path = Path(dest_path) / Path(obj_key).relative_to(key)
            dest_file_path.parent.mkdir(parents=True, exist_ok=True)
            self.client.download_file(bucket_name, obj_key, str(dest_file_path))

    def _cluster_to_local(self, cluster, dest_path):
        if not cluster.ips:
            raise ValueError("Cluster must be started before copying data from it.")

        Path(dest_path).expanduser().mkdir(parents=True, exist_ok=True)
        cluster.rsync(
            source=self.path,
            dest=str(Path(dest_path).expanduser()),
            up=False,
            contents=True,
        )
        new_folder = copy.deepcopy(self)
        new_folder.path = dest_path
        new_folder.system = "file"
        return new_folder

    def _move_within_s3(self, new_path):
        bucket_name = self._bucket_name
        key = self._key
        s3_objects = self.client.list_objects_v2(Bucket=bucket_name, Prefix=key)
        for obj in s3_objects.get("Contents", []):
            old_key = obj["Key"]
            new_key = new_path + old_key[len(key) :]
            self.client.copy_object(
                Bucket=bucket_name,
                CopySource={"Bucket": bucket_name, "Key": old_key},
                Key=new_key,
            )
            self.client.delete_object(Bucket=bucket_name, Key=old_key)

    def _s3_to_local(self, local_path):
        bucket_name = self._bucket_name
        key = self._key
        Path(local_path).mkdir(parents=True, exist_ok=True)
        s3_objects = self.client.list_objects_v2(Bucket=bucket_name, Prefix=key)
        for obj in s3_objects.get("Contents", []):
            obj_key = obj["Key"]
            dest_file_path = Path(local_path) / Path(obj_key).relative_to(key)
            dest_file_path.parent.mkdir(parents=True, exist_ok=True)
            self.client.download_file(bucket_name, obj_key, str(dest_file_path))
            self.client.delete_object(Bucket=bucket_name, Key=obj_key)

    def _s3_copy(self, new_path):
        bucket_name = self._bucket_name
        key = self._key
        s3_objects = self.client.list_objects_v2(Bucket=bucket_name, Prefix=key)
        for obj in s3_objects.get("Contents", []):
            old_key = obj["Key"]
            new_key = new_path + old_key[len(key) :]
            self.client.copy_object(
                Bucket=bucket_name,
                CopySource={"Bucket": bucket_name, "Key": old_key},
                Key=new_key,
            )

    def put(
        self,
        contents: Union["S3Folder", Dict],
        overwrite: bool = False,
        mode: str = "wb",
    ):
        """Put given contents in folder."""
        self.mkdir()
        if isinstance(contents, list):
            for resource in contents:
                self.put(resource, overwrite=overwrite)
            return

        key = self._key
        bucket_name = self._bucket_name

        if isinstance(contents, S3Folder):
            if contents.folder_path is None:
                contents.folder_path = key + "/" + contents.folder_path
            return

        if not isinstance(contents, Dict):
            raise TypeError(
                "`contents` argument must be a dict mapping filenames to file-like objects"
            )

        if overwrite is False:
            # Check if files exist and raise an error if they do
            existing_files = [
                obj["Key"]
                for obj in self.client.list_objects_v2(
                    Bucket=bucket_name, Prefix=key
                ).get("Contents", [])
            ]
            intersection = set(existing_files).intersection(set(contents.keys()))
            if intersection:
                raise FileExistsError(
                    f"File(s) {intersection} already exist(s) at path {key}, "
                    f"cannot save them without overwriting."
                )

        for filename, file_obj in contents.items():
            file_key = key + filename
            try:
                body = self._serialize_file_obj(file_obj)
                self.client.put_object(Bucket=bucket_name, Key=file_key, Body=body)

            except Exception as e:
                raise RuntimeError(f"Failed to upload {filename} to S3: {e}")

    def mv(self, system, path: Optional[str] = None):
        """Move the folder to a new filesystem or cluster."""
        if path is None:
            path = "rh/" + self.rns_address

        if system == "s3":
            self._move_within_s3(path)
        elif system == "file":
            self._s3_to_local(path)
        else:
            raise NotImplementedError("System not supported")

        self.path = path
        self.system = system

    def ls(self, full_paths: bool = True, sort: bool = False) -> List:
        """List the contents of the folder.

        Args:
            full_paths (Optional[bool]): Whether to list the full paths of the folder contents.
                Defaults to ``True``.
            sort (Optional[bool]): Whether to sort the folder contents by time modified.
                Defaults to ``False``.
        """
        paginator = self.client.get_paginator("list_objects_v2")
        pages = paginator.paginate(Bucket=self._bucket_name, Prefix=self._key)

        paths = []
        for page in pages:
            for obj in page.get("Contents", []):
                paths.append(obj)

        if sort:
            paths = sorted(paths, key=lambda f: f["LastModified"], reverse=True)

        if full_paths:
            return [
                self._urlpath + f"{self._bucket_name}/{obj['Key']}" for obj in paths
            ]
        else:
            return [Path(obj["Key"]).name for obj in paths]

    def exists_in_system(self):
        """Whether the folder exists in the filesystem."""
        try:
            # Check if there are any objects with the given prefix
            response = self.client.list_objects_v2(
                Bucket=self._bucket_name, Prefix=self._key, MaxKeys=1
            )
            return "Contents" in response
        except Exception as e:
            logger.error(f"Failed to check if folder exists: {e}")
            return False

    def open(self, name, mode="rb", encoding=None):
        """Returns an S3 object stream which must be used as a content manager to be opened.

        Example:
            >>> with my_folder.open('obj_name') as my_file:
            >>>     pickle.load(my_file)
        """
        obj_key = self._key + name
        bucket_name = self._bucket_name

        try:
            obj = self.client.get_object(Bucket=bucket_name, Key=obj_key)
            if "r" not in mode:
                raise NotImplementedError(f"{mode} mode is not implemented yet for S3")

            return obj["Body"]

        except Exception as e:
            raise e

    def mkdir(self):
        """Create the folder in specified file system if it doesn't already exist."""
        try:
            bucket = self._bucket_name
            key = self._key
            self.client.put_object(Bucket=bucket, Key=key)
            logger.info(
                f"Folder with path: {key} created successfully in bucket {bucket}."
            )
            return self

        except Exception as e:
            raise e

    def rm(self, contents: list = None, recursive: bool = True):
        """Delete a folder from the S3 bucket. Optionally provide a list of folder contents to delete.

        Args:
            contents (Optional[List]): Specific contents to delete in the folder.
            recursive (bool): Delete the folder itself (including all its contents).
                Defaults to ``True``.
        """
        key = self._key
        bucket = self._bucket_name
        if contents:
            objects_to_delete = [{"Key": f"{key}{content}"} for content in contents]
        else:
            if recursive:
                paginator = self.client.get_paginator("list_objects_v2")
                pages = paginator.paginate(Bucket=bucket, Prefix=key)
                objects_to_delete = [
                    {"Key": obj["Key"]}
                    for page in pages
                    for obj in page.get("Contents", [])
                ]
            else:
                objects_to_delete = [{"Key": key}]

        if objects_to_delete:
            self.client.delete_objects(
                Bucket=bucket, Delete={"Objects": objects_to_delete}
            )

    def delete_bucket(self):
        """Delete the s3 bucket."""
        try:
            from sky.data.storage import S3Store

            S3Store(name=self._bucket_name, source=self._urlpath).delete()
        except Exception as e:
            raise e

    def _upload(self, src: str, region: Optional[str] = None):
        """Upload a folder to an S3 bucket."""
        sync_dir_command = self._upload_command(src=src, dest=self.path)
        self._upload_folder_to_bucket(sync_dir_command)

    def _upload_command(self, src: str, dest: str):
        # https://github.com/skypilot-org/skypilot/blob/983f5fa3197fe7c4b5a28be240f7b027f7192b15/sky/data/storage.py#L922
        dest = dest.lstrip("/")
        upload_command = (
            'aws s3 sync --no-follow-symlinks --exclude ".git/*" '
            f"{src} "
            f"{self._urlpath}{dest}"
        )

        return upload_command

    def _download(self, dest):
        """Download a folder from an S3 bucket to local dir."""
        # NOTE: Sky doesn't support this API yet for each provider
        # https://github.com/skypilot-org/skypilot/blob/983f5fa3197fe7c4b5a28be240f7b027f7192b15/sky/data/storage.py#L231
        remote_dir = self.path.lstrip("/")
        remote_dir = self._urlpath + remote_dir
        try:
            subprocess.run(
                ["aws", "s3", "sync", remote_dir, dest],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                check=True,
            )
        except subprocess.CalledProcessError as e:
            raise e

    def _download_command(self, src: str, dest: str):
        # https://github.com/skypilot-org/skypilot/blob/983f5fa3197fe7c4b5a28be240f7b027f7192b15/sky/cloud_stores.py#L68
        download_via_awscli = "aws s3 sync --no-follow-symlinks " f"{src} {dest}"
        return download_via_awscli

    def _to_cluster(self, dest_cluster, path=None):
        """Copy the folder from a s3 bucket onto a cluster."""
        download_command = self._download_command(src=self.fsspec_url, dest=path)
        dest_cluster.run_bash([download_command])
        return S3Folder(path=path, system=dest_cluster, dryrun=True)

    def _to_local(self, dest_path: str):
        """Copy a folder from an S3 bucket to local dir."""
        self._download(dest=dest_path)
        return self._destination_folder(dest_path=dest_path, dest_system="file")

    def _to_data_store(
        self,
        system: str,
        data_store_path: Optional[str] = None,
    ):
        """Copy folder from S3 to another remote data store (ex: S3, GCP, Azure)"""
        if system == "s3":
            # Transfer between S3 folders
            sync_dir_command = self._upload_command(
                src=self.fsspec_url, dest=data_store_path
            )
            self._upload_folder_to_bucket(sync_dir_command)
        elif system == "gs":
            # Note: The sky data transfer API only allows for transfers between buckets, not specific directories.
            logger.warning(
                "Transfer from S3 to GCS currently supported for buckets only, not specific directories."
            )
            s3_bucket_name = self._bucket_name
            gs_bucket_name = self._bucket_name_from_path(data_store_path)
            self.s3_to_gcs(
                s3_bucket_name=s3_bucket_name,
                gs_bucket_name=gs_bucket_name,
            )
        elif system == "azure":
            raise NotImplementedError("Azure not yet supported")
        else:
            raise ValueError(f"Invalid system: {system}")

        return self._destination_folder(dest_path=data_store_path, dest_system=system)

    def s3_to_gcs(self, s3_bucket_name: str, gs_bucket_name: str):
        import boto3
        from google import auth
        from googleapiclient import discovery
        from oauth2client.client import GoogleCredentials

        from runhouse import GCSFolder

        # Adapted from:
        # https://github.com/skypilot-org/skypilot/blob/983f5fa3197fe7c4b5a28be240f7b027f7192b15/sky/data/data_transfer.py#L36

        oauth_credentials = GoogleCredentials.get_application_default()
        storagetransfer = discovery.build(
            "storagetransfer", "v1", credentials=oauth_credentials
        )

        try:
            credentials, project_id = auth.default()
        except Exception as e:
            raise e

        if project_id is None:
            raise ValueError(
                "Failed to get GCP project id. Please make sure you have "
                "run the following: gcloud init; "
                "Alternatively, set the project ID using: gcloud config set project <project_id> "
                "or with an environment variable: export GOOGLE_CLOUD_PROJECT=<project_id>"
            )

        session = boto3.Session()

        aws_credentials = session.get_credentials().get_frozen_credentials()

        storage_account = (
            storagetransfer.googleServiceAccounts().get(projectId=project_id).execute()
        )

        GCSFolder.add_bucket_iam_member(
            gs_bucket_name,
            "roles/storage.admin",
            "serviceAccount:" + storage_account["accountEmail"],
            project_id=project_id,
        )

        transfer_job = {
            "description": f"Transferring data from S3 Bucket \
            {s3_bucket_name} to GCS Bucket {gs_bucket_name}",
            "status": "ENABLED",
            "projectId": project_id,
            "transferSpec": {
                "awsS3DataSource": {
                    "bucketName": s3_bucket_name,
                    "awsAccessKey": {
                        "accessKeyId": aws_credentials.access_key,
                        "secretAccessKey": aws_credentials.secret_key,
                    },
                },
                "gcsDataSink": {
                    "bucketName": gs_bucket_name,
                },
            },
        }

        response = storagetransfer.transferJobs().create(body=transfer_job).execute()

        operation = (
            storagetransfer.transferJobs()
            .run(jobName=response["name"], body={"projectId": project_id})
            .execute()
        )
        logger.info(
            f"Transfer job scheduled: {self._urlpath}{s3_bucket_name} -> gs://{gs_bucket_name}"
        )

        logger.info("Waiting for the transfer to finish")

        timeout = False
        start = time.time()

        while True:
            # Get the status of the transfer operation
            result = (
                storagetransfer.transferOperations()
                .get(name=operation["name"])
                .execute()
            )

            if "error" in result:
                raise RuntimeError(result["error"])

            if "done" in result and result["done"]:
                logger.info(
                    f"Transfer finished in {(time.time() - start) / 60:.2f} minutes."
                )
                break

            # Check if the elapsed time exceeds the timeout
            if (time.time() - start) > TIMEOUT_SECONDS:
                timeout = True
                break

            time.sleep(POLL_INTERVAL)

        if timeout:
            logger.info(
                f"Transfer timed out after {(time.time() - start) / TIMEOUT_SECONDS:.2f} "
                "hours. Please check the status of the transfer job in the GCP "
                "Storage Transfer Service console at "
                "https://cloud.google.com/storage-transfer-service"
            )



import inspect
from typing import Any, Optional, Tuple

from runhouse.resources.module import Module, MODULE_ATTRS


class Asgi(Module):
    def __init__(self, app_pointers: Optional[Tuple] = None, **kwargs):
        """
        Runhouse Asgi Module. Allows you to deploy an ASGI app to a cluster.

        .. note::
                To create an Asgi, please use the factory method :func:`asgi`.
        """
        super().__init__(**kwargs)
        self.app_pointers = app_pointers
        self._app = None

    async def asgi_call(self, scope, receive, send) -> None:
        if not self._app:
            self._app = self._get_obj_from_pointers(*self.app_pointers)
        await self._app(scope, receive, send)

    def __call__(self, *args, **kwargs):
        return self.asgi_call(*args, **kwargs)

    def _route_names_and_endpoints(self):
        if not self._app:
            self.local._app = Module._get_obj_from_pointers(*self.local.app_pointers)

        return {route.name: route.endpoint for route in self.local._app.routes[4:]}

    def signature(self, rich=False):
        sig = super().signature(rich=rich)

        route_attrs = {
            name: self.method_signature(endpoint) if rich else None
            for name, endpoint in self._route_names_and_endpoints().items()
            if not name[0] == "_"
            and name not in MODULE_ATTRS
            and name not in dir(Module)
            and callable(endpoint)
            and not (
                "local" in endpoint.parameters and endpoint.parameters["local"].default
            )
        }
        sig.update(route_attrs)
        return sig

    def __getstate__(self):
        state = self.__dict__.copy()
        state["_app"] = None
        return state

    def _signature_extensions(self, item):
        if item == "_app":
            # Object hasn't been remotely initialized yet
            return None

        # If the item is a route name, return the endpoint. This allows us to call the route name directly
        # as a method, which is nice for debugging.
        return self._route_names_and_endpoints().get(item)


def asgi(
    app: Any,
    name: Optional[str] = None,
    load_from_den: bool = True,
    dryrun: bool = False,
    **kwargs
) -> Asgi:
    """
    A factory method for creating ASGI modules, such as a FastAPI or Starlette app.

    Args:
        app (Any): The ASGI app to deploy.
        name (Optional[str], optional): Name to give to the module resource, if you wish to rename it.
            (Default: ``None``)
        load_from_den (bool): Whether to try loading the Cluster resource from Den. (Default: ``True``)
        dryrun (bool): Whether to create the Cluster if it doesn't exist, or load a Cluster object as a dryrun.
            (Default: ``False``)
        **kwargs: Keyword arguments for the :class:`Asgi` constructor.

    Returns:
        Asgi: The resulting Asgi object.

    Example:
        >>> from fastapi import FastAPI
        >>> import runhouse as rh
        >>> app = FastAPI()
        >>> @app.get("/summer/{a}")
        >>> def summer(a: int, b: int = 2):
        >>>     return a + b
        >>>
        >>> fast_api_module = rh.asgi(app).to(rh.here, name="fast_api_module")
        >>> fast_api_module.summer(1, 2)
        >>> # output: 3
    """
    if name and not any([app, kwargs]):
        # Try reloading existing module
        return Asgi.from_name(name, load_from_den=load_from_den, dryrun=dryrun)

    callers_global_vars = inspect.currentframe().f_back.f_globals.items()
    app_name = [
        var_name for var_name, var_val in callers_global_vars if var_val is app
    ][0]

    app_file, app_module, _ = Module._extract_pointers(app.routes[4].endpoint)

    return Asgi(
        app_pointers=(app_file, app_module, app_name),
        name=name,
        dryrun=dryrun,
        **kwargs
    )



import copy
import json
import os
from pathlib import Path
from typing import Any, Dict, Optional, Union

from runhouse.constants import DEFAULT_PROCESS_NAME
from runhouse.globals import configs, rns_client
from runhouse.resources.hardware.cluster import Cluster
from runhouse.resources.hardware.utils import _get_cluster_from
from runhouse.resources.secrets.secret import Secret
from runhouse.resources.secrets.utils import _check_file_for_mismatches
from runhouse.utils import create_local_dir


class ProviderSecret(Secret):
    _PROVIDER = None
    _DEFAULT_CREDENTIALS_PATH = None
    _DEFAULT_ENV_VARS = {}

    def __init__(
        self,
        name: Optional[str] = None,
        provider: Optional[str] = None,
        values: Dict = None,
        path: str = None,
        env_vars: Dict = None,
        dryrun: bool = False,
        **kwargs,
    ):
        """
        Provider Secret class. Built-in provider classes contain default path and/or environment variable mappings,
        based on it's expected usage.

        Currently supported built-in providers:
        anthropic, aws, azure, gcp, github, huggingface, lambda, langchain, openai, pinecone, ssh, sky, wandb.

        .. note::
            To create a ProviderSecret, please use the factory method :func:`provider_secret`.
        """
        super().__init__(name=name, values=values, dryrun=dryrun)
        self.provider = provider or self._PROVIDER
        self.path = path
        self.env_vars = env_vars

        if not any([values, path, env_vars]):
            if self._from_path(self._DEFAULT_CREDENTIALS_PATH):
                self.path = self._DEFAULT_CREDENTIALS_PATH
            elif self._from_env(self._DEFAULT_ENV_VARS):
                self.env_vars = self._DEFAULT_ENV_VARS
            else:
                raise ValueError(
                    "Secrets values not provided and could not be extracted from default file "
                    f"({self._DEFAULT_CREDENTIALS_PATH}) or env vars ({self._DEFAULT_ENV_VARS.values()}) locations."
                )

    @property
    def values(self):
        if self._values:
            return self._values
        elif self.path:
            return self._from_path(self.path)
        elif self.env_vars:
            return self._from_env(self.env_vars)
        return {}

    def config(self, condensed: bool = True, values: bool = True):
        config = super().config(condensed)
        config.update({"provider": self.provider})
        if self.path:
            config.update({"path": self.path})
        if self.env_vars:
            config.update({"env_vars": self.env_vars})
        if not values:
            config.pop("values", None)
        return config

    @staticmethod
    def from_config(config: dict, dryrun: bool = False, _resolve_children: bool = True):
        """Create a ProviderSecret object from a config dictionary."""
        return ProviderSecret(**config, dryrun=dryrun)

    def save(
        self,
        name: str = None,
        save_values: bool = True,
        headers: Optional[Dict] = None,
        folder: str = None,
    ):
        name = name or self.rns_address or self.name or self.provider
        return super().save(
            name=name, save_values=save_values, headers=headers, folder=folder
        )

    def delete(self, headers: Optional[Dict] = None, contents: bool = False):
        """Delete the secret config from Den and from Vault/local. Optionally also delete contents of secret file
        or env vars."""
        headers = headers or rns_client.request_headers()
        if self.path and contents and os.path.exists(os.path.expanduser(self.path)):
            os.remove(os.path.expanduser(self.path))
        elif self.env_vars and contents:
            for (_, env_var) in self.env_vars.keys():
                if env_var in os.environ:
                    del os.environ[env_var]
        super().delete(headers=headers)

    def write(
        self,
        path: str = None,
        env_vars: Dict = None,
        file: bool = False,
        env: bool = False,
        overwrite: bool = False,
        write_config: bool = True,
    ):
        if not self.values:
            raise ValueError("Could not determine values to write down.")
        if (file or path) and (env or env_vars):
            raise ValueError("Can only save to one of file or env at a given time.")
        if not any([file, env, path, env_vars]):
            file = True  # default write to file

        if file or path:
            path = path or self.path or self._DEFAULT_CREDENTIALS_PATH
            return self._write_to_file(
                path, values=self.values, overwrite=overwrite, write_config=write_config
            )
        elif env or env_vars:
            env_vars = env_vars or self.env_vars or self._DEFAULT_ENV_VARS
            return self._write_to_env(env_vars, values=self.values, overwrite=overwrite)

    def to(
        self,
        system: Union[str, Cluster],
        path: str = None,
        process: Optional[str] = None,
        values: bool = None,
        name: Optional[str] = None,
    ):
        """Return a copy of the secret on a system.

        Args:
            system (str or Cluster): Cluster to send the secret to
            path (str or Path, optional): Path on cluster to write down the secret values to.
                If not provided and secret is not already associated with a path, the secret values
                will not be written down on the cluster.
            process (str, optional): Process on the cluster to send the secret to, to set the secret env var
                values inside the process.
            values (bool, optional): Whether to save down the values in the resource config. By default,
                save down values if the secret is not being written down to a file or environment variable.
                Otherwise, values are not written down. (Default: None)
            name (str, ooptional): Name to assign the resource on the cluster.

        Example:
            >>> secret.to(my_cluster, path=secret.path)
        """

        system = _get_cluster_from(system)
        path = path or self.path

        if system.on_this_cluster():
            if not process and not path == self.path:
                if name and not self.name == name:
                    self.rename(name)
                return self
            self.write(path=path, env=process)
            if path and len(system.ips) > 1:
                for node in system.ips[1:]:
                    system._local_rsync(src=path, dest=Path(path).parent, node=node)
            new_secret = copy.deepcopy(self)
            new_secret._values = None
            new_secret.path = path
            new_secret.name = name or self.name
            return new_secret

        new_secret = copy.deepcopy(self)
        new_secret.name = name or self.name or self.provider

        if values:
            new_secret._values = self.values
        elif values is None and not (path or process or self.env_vars):
            new_secret._values = self.values
        elif values is False:
            new_secret._values = None

        process = process or DEFAULT_PROCESS_NAME
        key = system.put_resource(new_secret, process=process)
        if path:
            new_secret.path = self._file_to(
                key=key, system=system, path=path, values=self.values
            )
            if len(system.ips) > 1:
                for node in system.ips[1:]:
                    system._local_rsync(source=path, dest=Path(path).parent, node=node)
        if process or self.env_vars:
            env_vars = self.env_vars or self._DEFAULT_ENV_VARS
            if env_vars:
                env_vars = self._map_env_vars(env_vars)
                system.set_env_vars_globally(env_vars=env_vars)
        return new_secret

    def _map_env_vars(self, env_vars: Dict = None):
        env_vars = env_vars or self.env_vars or self._DEFAULT_ENV_VARS
        mapped_env_vars = {
            env_vars[k]: self.values[k] for k in self.values if k in env_vars
        }
        return mapped_env_vars

    def _file_to(
        self,
        key: str,
        system: Union[str, Cluster],
        path: str = None,
        values: Any = None,
    ):
        system.call(key, "_write_to_file", path=path, values=values)
        return path

    def _write_to_file(
        self, path: str, values: Any, overwrite: bool = False, write_config: bool = True
    ):
        new_secret = copy.deepcopy(self)
        if not _check_file_for_mismatches(
            path, self._from_path(path), values, overwrite
        ):
            full_path = create_local_dir(path)
            with open(full_path, "w") as f:
                json.dump(values, f, indent=4)

            if write_config:
                self._add_to_rh_config(path)

        new_secret._values = None
        new_secret.path = path
        return new_secret

    def _write_to_env(self, env_vars: Dict, values: Any, overwrite: bool):
        existing_keys = dict(os.environ).keys()
        added_keys = []
        for key in env_vars.keys():
            if env_vars[key] not in existing_keys or overwrite:
                os.environ[env_vars[key]] = values[key]
                added_keys.append(env_vars[key])

        if added_keys:
            self._add_to_rh_config(added_keys)

        new_secret = copy.deepcopy(self)
        new_secret._values = None
        new_secret.env_vars = env_vars
        return new_secret

    def _from_env(self, env_vars: Dict = None):
        env_vars = env_vars or self.env_vars
        if not env_vars:
            return {}

        values = {}

        for key in env_vars.keys():
            try:
                values[key] = os.environ[env_vars[key]]
            except KeyError:
                return {}
        return values

    def _from_path(self, path: str = None):
        path = path or self.path
        if not path:
            return {}

        path = os.path.expanduser(path)
        if os.path.exists(path):
            with open(path) as f:
                try:
                    contents = json.load(f)
                except json.decoder.JSONDecodeError:
                    contents = f.read()
                return contents
        return {}

    @staticmethod
    def extract_secrets_from_path(path: str) -> Union[str, None]:
        secret_path = os.path.expanduser(path)

        if not os.path.exists(secret_path):
            return None

        provider_secret = Path(secret_path).read_text()

        return provider_secret

    def _add_to_rh_config(self, val):
        if not self.name:
            self.name = self.provider
        configs.set_nested(key="secrets", value={self.name: val})



import configparser
import copy
import os

from typing import Dict

from runhouse.resources.secrets.provider_secrets.provider_secret import ProviderSecret
from runhouse.resources.secrets.utils import _check_file_for_mismatches
from runhouse.utils import create_local_dir


class AWSSecret(ProviderSecret):
    """
    .. note::
            To create an AWSSecret, please use the factory method :func:`provider_secret` with ``provider="aws"``.
    """

    _PROVIDER = "aws"
    _DEFAULT_CREDENTIALS_PATH = "~/.aws/credentials"
    _DEFAULT_ENV_VARS = {
        "access_key": "AWS_ACCESS_KEY_ID",
        "secret_key": "AWS_SECRET_ACCESS_KEY",
    }

    @staticmethod
    def from_config(config: dict, dryrun: bool = False, _resolve_children: bool = True):
        return AWSSecret(**config, dryrun=dryrun)

    def _write_to_file(
        self,
        path: str,
        values: Dict,
        overwrite: bool = False,
        write_config: bool = True,
    ):
        new_secret = copy.deepcopy(self)

        if not _check_file_for_mismatches(
            path, self._from_path(path), values, overwrite
        ):

            parser = configparser.ConfigParser()
            section_name = "default"
            parser.add_section(section_name)
            parser.set(
                section=section_name,
                option="aws_access_key_id",
                value=values["access_key"],
            )
            parser.set(
                section=section_name,
                option="aws_secret_access_key",
                value=values["secret_key"],
            )

            full_path = create_local_dir(path)
            with open(full_path, "w+") as f:
                parser.write(f)

            if write_config:
                new_secret._add_to_rh_config(path)

        new_secret._values = None
        new_secret.path = path
        return new_secret

    def _from_path(self, path: str):
        config = configparser.ConfigParser()
        if path and os.path.exists(os.path.expanduser(path)):
            config.read(os.path.expanduser(path))
        else:
            return {}

        section_name = "default"
        access_key = config[section_name]["aws_access_key_id"]
        secret_key = config[section_name]["aws_secret_access_key"]

        return {
            "access_key": access_key,
            "secret_key": secret_key,
        }



import datetime
import ipaddress
from pathlib import Path

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID

from runhouse.logger import get_logger

from runhouse.rns.utils.api import resolve_absolute_path

logger = get_logger(__name__)


class TLSCertConfig:
    """Handler for creating and managing the TLS certs needed to enable HTTPS on the Runhouse API server."""

    CERT_NAME = "rh_server.crt"
    PRIVATE_KEY_NAME = "rh_server.key"
    TOKEN_VALIDITY_DAYS = 365

    # Base directory for certs stored locally
    # Note: Each user will have their own set of certs, to be reused on for each cluster
    LOCAL_CERT_DIR = "~/.rh/certs"

    # https://caddy.community/t/permission-denied-error-when-caddy-try-to-save-the-certificate/15026
    # Note: When running as a systemd service, Caddy as runs as the "caddy" user and doesn’t have permission to
    # read files in the home directory. Easiest solution to ensure there are no permission errors is to save the
    # certs in the Caddy cluster directory when Caddy is enabled.
    CADDY_CLUSTER_DIR = "/var/lib/caddy"
    DEFAULT_CLUSTER_DIR = "~/certs"

    def __init__(self, cert_path: str = None, key_path: str = None):
        self._cert_path = cert_path
        self._key_path = key_path

    @property
    def cert_path(self):
        if self._cert_path is not None:
            return resolve_absolute_path(self._cert_path)

        return str(Path(f"{self.LOCAL_CERT_DIR}/{self.CERT_NAME}").expanduser())

    @cert_path.setter
    def cert_path(self, cert_path):
        self._cert_path = cert_path

    @property
    def key_path(self):
        if self._key_path is not None:
            return resolve_absolute_path(self._key_path)

        return str(Path(f"{self.LOCAL_CERT_DIR}/{self.PRIVATE_KEY_NAME}").expanduser())

    @key_path.setter
    def key_path(self, key_path):
        self._key_path = key_path

    @property
    def cert_dir(self):
        return Path(self.cert_path).parent

    @property
    def key_dir(self):
        return Path(self.key_path).parent

    def generate_certs(self, address: str = None, domain: str = None):
        """Create a self-signed SSL certificate. This won't be verified by a CA, but the connection will
        still be encrypted. If domain is provided, prioritize using it to allow the same cert to be re-used
        across multiple servers. If not, use the provided address (IP or DNS name)"""
        # Generate the private key
        private_key = rsa.generate_private_key(
            public_exponent=65537, key_size=2048, backend=default_backend()
        )

        # Create a self-signed certificate
        subject = issuer = x509.Name(
            [x509.NameAttribute(NameOID.COMMON_NAME, "run.house")]
        )

        subject_names = [
            x509.DNSName("localhost"),
            x509.IPAddress(ipaddress.IPv4Address("127.0.0.1")),
        ]

        if domain:
            # Add the provided domain name to the SAN
            subject_names.append(x509.DNSName(domain))

        elif address:
            try:
                # Check if the address is a valid IP address and add it to the SAN
                ip_addr = ipaddress.IPv4Address(address)
                subject_names.append(x509.IPAddress(ip_addr))
            except ipaddress.AddressValueError:
                # If not a valid IP address, treat it as an additional DNS name
                subject_names.append(x509.DNSName(address))
        else:
            raise ValueError(
                "At least one of `address` or `domain` must be provided to generate certs."
            )

        # Add Subject Alternative Name extension
        san = x509.SubjectAlternativeName(subject_names)

        cert = (
            x509.CertificateBuilder()
            .subject_name(subject)
            .issuer_name(issuer)
            .public_key(private_key.public_key())
            .serial_number(x509.random_serial_number())
            .not_valid_before(datetime.datetime.utcnow())
            .not_valid_after(
                datetime.datetime.utcnow()
                + datetime.timedelta(days=self.TOKEN_VALIDITY_DAYS)
            )
            .add_extension(san, critical=False)
            .sign(private_key, hashes.SHA256(), default_backend())
        )

        self._write_cert_files(private_key, cert)

    def _write_cert_files(self, private_key, cert):
        """Save the private key and cert files on the system (either locally or on cluster)."""
        # Ensure the directories exist and have the correct permissions
        self.cert_dir.expanduser().mkdir(parents=True, mode=0o750, exist_ok=True)
        self.key_dir.expanduser().mkdir(parents=True, mode=0o750, exist_ok=True)

        with open(self.key_path, "wb") as key_file:
            key_file.write(
                private_key.private_bytes(
                    encoding=serialization.Encoding.PEM,
                    format=serialization.PrivateFormat.TraditionalOpenSSL,
                    encryption_algorithm=serialization.NoEncryption(),
                )
            )

        # Save the certificate to a file (server.crt)
        with open(self.cert_path, "wb") as cert_file:
            cert_file.write(cert.public_bytes(encoding=serialization.Encoding.PEM))

        logger.info(
            f"Certificate and private key files saved in paths: {self.cert_path} and {self.key_path}"
        )



import os
import sys
from pathlib import Path
from typing import List

RESERVED_SYSTEM_NAMES: List[str] = ["file", "s3", "gs", "azure", "here"]
CLUSTER_CONFIG_PATH: str = "~/.rh/cluster_config.json"
CONFIG_YAML_PATH: str = "~/.rh/config.yaml"
SERVER_LOGFILE_PATH = "~/.rh/server.log"
LOCALHOST: str = "127.0.0.1"
LOCAL_HOSTS: List[str] = ["localhost", LOCALHOST]
TUNNEL_TIMEOUT = 5
NUM_PORTS_TO_TRY = 10

LOGS_DIR = ".rh/logs"
RH_LOGFILE_PATH = Path.home() / LOGS_DIR

ENVS_DIR = "~/.rh/envs"

GIGABYTE = 1024**3
MAX_MESSAGE_LENGTH = 1 * GIGABYTE


CLI_RESTART_CMD = "runhouse server restart"
CLI_START_CMD = "runhouse server start"
CLI_STOP_CMD = "runhouse server stop"

DEFAULT_SERVER_PORT = 32300
DEFAULT_HTTPS_PORT = 443
DEFAULT_HTTP_PORT = 80
DEFAULT_SSH_PORT = 22

DEFAULT_RAY_PORT = 6379
DEFAULT_DASK_PORT = 8786


DEFAULT_SERVER_HOST = "0.0.0.0"

LOGGING_WAIT_TIME = 0.5
LOGS_TO_SHOW_UP_CHECK_TIME = 0.1
MAX_LOGS_TO_SHOW_UP_WAIT_TIME = 5

# Commands
SERVER_START_CMD = f"{sys.executable} -m runhouse.servers.http.http_server"
SERVER_STOP_CMD = f'pkill -f "{SERVER_START_CMD}"'
# 2>&1 redirects stderr to stdout
SERVER_LOGFILE = os.path.expanduser("~/.rh/server.log")
START_SCREEN_CMD = (
    f"screen -dm bash -c \"{SERVER_START_CMD} 2>&1 | tee -a '{SERVER_LOGFILE}' 2>&1\""
)
START_NOHUP_CMD = f"nohup {SERVER_START_CMD} >> {SERVER_LOGFILE} 2>&1 &"
# We need to specify "--disable-usage-stats" because cluster.run uses INTERACTIVE (-t) SshMode by
# default, which Ray detects and asks the user for 10 seconds whether they want to opt out of usage
# stats collection. This breaks the daemon start sequence, so we disable it upfront.
RAY_START_CMD = f"ray start --head --port {DEFAULT_RAY_PORT} --disable-usage-stats"
# RAY_BOOTSTRAP_FILE = "~/ray_bootstrap_config.yaml"
# --autoscaling-config=~/ray_bootstrap_config.yaml
# We need to use this instead of ray stop to make sure we don't stop the SkyPilot ray server,
# which runs on other ports but is required to preserve autostop and correct cluster status.
RAY_KILL_CMD = 'pkill -f ".*ray.*' + str(DEFAULT_RAY_PORT) + '.*"'

CONDA_INSTALL_CMDS = [
    "wget https://repo.continuum.io/miniconda/Miniconda3-latest-Linux-x86_64.sh -O ~/miniconda.sh",
    "bash ~/miniconda.sh -b -p ~/miniconda",
    "source $HOME/miniconda/bin/activate",
]
# TODO should default to user's local Python version?
# from platform import python_version; python_version()
CONDA_PREFERRED_PYTHON_VERSION = "3.10.9"

DEFAULT_PROCESS_NAME = "default_process"
DEFAULT_DOCKER_CONTAINER_NAME = "sky_container"

# Constants for the status check
DOUBLE_SPACE_UNICODE = "\u00A0\u00A0"
BULLET_UNICODE = "\u2022"
SECOND = 1
MINUTE = 60
HOUR = 3600
DAY = HOUR * 24
DEFAULT_STATUS_CHECK_INTERVAL = 1 * MINUTE
DEFAULT_AUTOSTOP_CHECK_INTERVAL = 1 * MINUTE
INCREASED_STATUS_CHECK_INTERVAL = 5 * MINUTE
GPU_COLLECTION_INTERVAL = 5 * SECOND
INCREASED_GPU_COLLECTION_INTERVAL = 1 * MINUTE

# We collect gpu every GPU_COLLECTION_INTERVAL.
# Meaning that in one minute we collect (MINUTE / GPU_COLLECTION_INTERVAL) gpu stats.
# Currently, we save gpu info of the last 10 minutes or less.
MAX_GPU_INFO_LEN = (MINUTE / GPU_COLLECTION_INTERVAL) * 10

# If we just collect the gpu stats (and not send them to den), the gpu_info dictionary *will not* be reseted by the servlets.
# Therefore, we need to cut the gpu_info size, so it doesn't consume too much cluster memory.
# Currently, we reduce the size by half, meaning we only keep the gpu_info of the last (MAX_GPU_INFO_LEN / 2) minutes.
REDUCED_GPU_INFO_LEN = MAX_GPU_INFO_LEN / 2

DEFAULT_LOG_LEVEL = "INFO"
# Surfacing Logs to Den constants
DEFAULT_LOG_SURFACING_INTERVAL = 2 * MINUTE
SERVER_LOGS_FILE_NAME = "server.log"
DEFAULT_SURFACED_LOG_LENGTH = 20
# Constants for schedulers
INCREASED_INTERVAL = 1 * HOUR

# Constants runhouse cluster list
LAST_ACTIVE_AT_TIMEFRAME = 24 * HOUR
MAX_CLUSTERS_DISPLAY = 200

# in seconds
TIME_UNITS = {"s": SECOND, "m": MINUTE, "h": HOUR, "d": DAY}
# ANSI Constants
ITALIC_BOLD = "\x1B[3m\x1B[1m"
RESET_FORMAT = "\x1B[0m"

SKY_VENV = "~/skypilot-runtime"

SSH_SKY_SECRET_NAME = "ssh-sky-key"



import copy
import shutil
import subprocess
from pathlib import Path
from typing import Dict, List, Optional

from runhouse.logger import get_logger

from .folder import Folder

logger = get_logger(__name__)


class GCSFolder(Folder):
    RESOURCE_TYPE = "folder"
    DEFAULT_FS = "gcp"

    def __init__(self, dryrun: bool, **kwargs):
        try:
            from google.cloud import storage
        except ImportError:
            raise ImportError(
                "`google-cloud-storage` is required for GCS folders. "
                "You can install it with `pip install google-cloud-storage`."
            )

        super().__init__(dryrun=dryrun, **kwargs)
        self.client = storage.Client()
        self._urlpath = "gs://"

    @staticmethod
    def from_config(config: Dict, dryrun: bool = False, _resolve_children: bool = True):
        """Load config values into the object."""
        return GCSFolder(**config, dryrun=dryrun)

    @property
    def bucket(self):
        return self.client.bucket(self._bucket_name)

    def _to_local(self, dest_path: str):
        """Copies folder to local."""
        from runhouse import Cluster

        if self._fs_str == "file":
            shutil.copytree(src=self.path, dst=dest_path)
        elif isinstance(self.system, Cluster):
            return self._cluster_to_local(cluster=self.system, dest_path=dest_path)
        else:
            self._gcs_copy_to_local(dest_path)

        return self._destination_folder(dest_path=dest_path, dest_system="file")

    def _gcs_copy_to_local(self, dest_path: str):
        """Copy GCS folder to local."""
        Path(dest_path).mkdir(parents=True, exist_ok=True)
        key = self._key
        blobs = self.client.list_blobs(self.bucket.name, prefix=key)
        for blob in blobs:
            dest_file_path = Path(dest_path) / Path(blob.name).relative_to(key)
            dest_file_path.parent.mkdir(parents=True, exist_ok=True)
            blob.download_to_filename(str(dest_file_path))

    def _cluster_to_local(self, cluster, dest_path):
        if not cluster.ips:
            raise ValueError("Cluster must be started before copying data from it.")
        Path(dest_path).expanduser().mkdir(parents=True, exist_ok=True)

        cluster.rsync(
            source=self.path,
            dest=str(Path(dest_path).expanduser()),
            up=False,
            contents=True,
        )
        new_folder = copy.deepcopy(self)
        new_folder.path = dest_path
        new_folder.system = "file"
        return new_folder

    def _move_within_gcs(self, new_path):
        key = self._key
        blobs = self.client.list_blobs(self.bucket.name, prefix=key)
        for blob in blobs:
            old_name = blob.name
            new_name = new_path + old_name[len(key) :]
            self.bucket.copy_blob(blob, self.bucket, new_name)
            blob.delete()

    def _gcs_to_local(self, local_path):
        key = self._key
        Path(local_path).mkdir(parents=True, exist_ok=True)
        blobs = self.client.list_blobs(self.bucket.name, prefix=key)
        for blob in blobs:
            dest_file_path = Path(local_path) / Path(blob.name).relative_to(key)
            dest_file_path.parent.mkdir(parents=True, exist_ok=True)
            blob.download_to_filename(str(dest_file_path))
            blob.delete()

    def _gcs_copy(self, new_path):
        key = self._key
        blobs = self.client.list_blobs(self.bucket.name, prefix=key)
        for blob in blobs:
            old_name = blob.name
            new_name = new_path + old_name[len(key) :]
            new_blob = self.bucket.blob(new_name)
            new_blob.rewrite(blob)

    def put(self, contents, overwrite=False, mode: str = "wb"):
        """Put given contents in folder."""
        self.mkdir()
        if isinstance(contents, list):
            for resource in contents:
                self.put(resource, overwrite=overwrite)
            return

        key = self._key

        if isinstance(contents, GCSFolder):
            if contents.folder_path is None:
                contents.folder_path = key + "/" + contents.folder_path
            return

        if not isinstance(contents, dict):
            raise TypeError(
                "`contents` argument must be a dict mapping filenames to file-like objects"
            )

        if overwrite is False:
            # Check if files exist and raise an error if they do
            existing_files = [
                blob.name for blob in self.client.list_blobs(self.bucket, prefix=key)
            ]
            intersection = set(existing_files).intersection(set(contents.keys()))
            if intersection:
                raise FileExistsError(
                    f"File(s) {intersection} already exist(s) at path {key}, "
                    f"cannot save them without overwriting."
                )

        for filename, file_obj in contents.items():
            file_key = key + filename
            try:
                blob = self.bucket.blob(file_key)
                file_obj = self._serialize_file_obj(file_obj)
                blob.upload_from_file(file_obj)

            except Exception as e:
                raise RuntimeError(f"Failed to upload {filename} to GCS: {e}")

    def mv(self, system, path: Optional[str] = None):
        """Move the folder to a new filesystem or cluster."""
        if path is None:
            path = "rh/" + self.rns_address

        if system == "gcs":
            self._move_within_gcs(path)
        elif system == "file":
            self._gcs_to_local(path)
        else:
            raise NotImplementedError("System not supported")

        self.path = path
        self.system = system

    def ls(self, full_paths: bool = True, sort: bool = False) -> List:
        """List the contents of the folder.

        Args:
            full_paths (Optional[bool]): Whether to list the full paths of the folder contents.
                Defaults to ``True``.
            sort (Optional[bool]): Whether to sort the folder contents by time modified.
                Defaults to ``False``.
        """
        blobs = list(self.client.list_blobs(self.bucket, prefix=self._key))

        if sort:
            blobs = sorted(blobs, key=lambda f: f.updated, reverse=True)

        if full_paths:
            return [self._urlpath + f"{self.bucket.name}/{blob.name}" for blob in blobs]
        else:
            return [Path(blob.name).name for blob in blobs]

    def exists_in_system(self):
        """Whether the folder exists in the filesystem."""
        try:
            blobs = list(
                self.client.list_blobs(self.bucket, prefix=self._key, max_results=1)
            )
            return len(blobs) > 0

        except Exception as e:
            logger.error(f"Failed to check if folder exists: {e}")
            return False

    def open(self, name, mode="rb", encoding=None):
        """Returns a GCS blob stream which must be used as a content manager to be opened.

        Example:
            >>> with my_folder.open('obj_name') as my_file:
            >>>     pickle.load(my_file)
        """
        blob_name = self._key + name
        blob = self.bucket.blob(blob_name)

        try:
            if "r" not in mode:
                raise NotImplementedError(f"{mode} mode is not implemented yet for GCS")

            return blob.open(mode=mode, encoding=encoding)

        except Exception as e:
            raise e

    def mkdir(self):
        """Create the folder in specified file system if it doesn't already exist."""
        try:
            key = self._key
            blob = self.bucket.blob()
            blob.upload_from_string("")
            logger.info(
                f"Directory {key} created successfully in bucket {self._bucket_name}."
            )
            return self
        except Exception as e:
            raise e

    def rm(self, contents: list = None, recursive: bool = True):
        """Delete a folder from the GCS bucket. Optionally provide a list of folder contents to delete.

        Args:
            contents (Optional[List]): Specific contents to delete in the folder.
            recursive (bool): Delete the folder itself (including all its contents).
                Defaults to ``True``.
        """
        key = self._key
        bucket = self._bucket_name
        if contents:
            blobs_to_delete = [
                self.bucket.blob(f"{key}{content}") for content in contents
            ]
        else:
            if recursive:
                blobs_to_delete = list(self.client.list_blobs(bucket, prefix=key))
            else:
                blobs_to_delete = [self.bucket.blob(key)]

        for blob in blobs_to_delete:
            blob.delete()

    def delete_bucket(self):
        """Delete the gcs bucket."""
        # https://github.com/skypilot-org/skypilot/blob/3517f55ed074466eadd4175e152f68c5ea3f5f4c/sky/data/storage.py#L1775
        bucket_name = self._bucket_name
        remove_obj_command = f"rm -r {self._urlpath}{bucket_name}"

        try:
            subprocess.check_output(
                remove_obj_command,
                stderr=subprocess.STDOUT,
                shell=True,
                executable="/bin/bash",
            )
        except subprocess.CalledProcessError as e:
            raise e

    def _upload(self, src: str, region: Optional[str] = None):
        """Upload a folder to an GCS bucket."""
        sync_dir_command = self._upload_command(src=src, dest=self.path)
        self._upload_folder_to_bucket(sync_dir_command)

    def _upload_command(self, src: str, dest: str):
        # https://github.com/skypilot-org/skypilot/blob/983f5fa3197fe7c4b5a28be240f7b027f7192b15/sky/data/storage.py#L1240
        dest = dest.lstrip("/")
        return f"gsutil -m rsync -r -x '.git/*' {src} {self._urlpath}{dest}"

    def _download(self, dest):
        """Download a folder from a GCS bucket to local dir."""
        # NOTE: Sky doesn't support this API yet for each provider
        # https://github.com/skypilot-org/skypilot/blob/983f5fa3197fe7c4b5a28be240f7b027f7192b15/sky/data/storage.py#L231
        remote_dir = self.path.lstrip("/")
        remote_dir = f"{self._urlpath}{remote_dir}"
        subprocess.run(
            ["gsutil", "-m", "rsync", "-r", "-x", ".git/*", remote_dir, dest],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            check=True,
        )

    def _download_command(self, src, dest):
        # https://github.com/skypilot-org/skypilot/blob/3517f55ed074466eadd4175e152f68c5ea3f5f4c/sky/cloud_stores.py#L139
        download_via_gsutil = f"gsutil -m rsync -e -r {src} {dest}"
        return download_via_gsutil

    def _to_cluster(self, dest_cluster, path=None):
        upload_command = self._upload_command(src=self.path, dest=path)
        dest_cluster.run_bash([upload_command])
        return GCSFolder(path=path, system=dest_cluster, dryrun=True)

    def _to_local(self, dest_path: str):
        """Copy a folder from an GCS bucket to local dir."""
        self._download(dest=dest_path)
        return self._destination_folder(dest_path=dest_path, dest_system="file")

    def _to_data_store(
        self,
        system: str,
        data_store_path: Optional[str] = None,
    ):
        """Copy folder from GCS to another remote data store (ex: GCS, S3, Azure)"""
        if system == "gs":
            # Transfer between GCS folders
            sync_dir_command = self._upload_command(
                src=self.fsspec_url, dest=data_store_path
            )
            self._upload_folder_to_bucket(sync_dir_command)
        elif system == "s3":
            # Note: The sky data transfer API only allows for transfers between buckets, not specific directories.
            logger.warning(
                "Transfer from GCS to S3 currently supported for buckets only, not specific directories."
            )
            gs_bucket_name = self._bucket_name
            s3_bucket_name = self._bucket_name_from_path(data_store_path)
            self.gcs_to_s3(
                gs_bucket_name=gs_bucket_name,
                s3_bucket_name=s3_bucket_name,
            )
        elif system == "azure":
            raise NotImplementedError("Azure not yet supported")
        else:
            raise ValueError(f"Invalid system: {system}")

        return self._destination_folder(dest_path=data_store_path, dest_system=system)

    def gcs_to_s3(self, gs_bucket_name: str, s3_bucket_name: str) -> None:
        # https://github.com/skypilot-org/skypilot/blob/3517f55ed074466eadd4175e152f68c5ea3f5f4c/sky/data/data_transfer.py#L138
        disable_multiprocessing_flag = '-o "GSUtil:parallel_process_count=1"'
        sync_command = f"gsutil -m {disable_multiprocessing_flag} rsync -rd {self._urlpath}{gs_bucket_name} s3://{s3_bucket_name}"
        try:
            subprocess.run(sync_command, shell=True)
        except subprocess.CalledProcessError as e:
            raise e

    @staticmethod
    def add_bucket_iam_member(
        bucket_name: str, role: str, member: str, project_id: str
    ) -> None:
        # https://github.com/skypilot-org/skypilot/blob/983f5fa3197fe7c4b5a28be240f7b027f7192b15/sky/data/data_transfer.py#L132
        from google.cloud import storage

        storage_client = storage.Client(project=project_id)
        bucket = storage_client.bucket(bucket_name)

        policy = bucket.get_iam_policy(requested_policy_version=3)
        policy.bindings.append({"role": role, "members": {member}})

        bucket.set_iam_policy(policy)

        logger.debug(f"Added {member} with role {role} to {bucket_name}.")



"""
Pulled from: https://github.com/skypilot-org/skypilot/blob/3baa9c9/sky/skylet/subprocess_daemon.py

Sky subprocess daemon.

Wait for parent_pid to exit, then SIGTERM (or SIGKILL if needed) the child
processes of proc_pid.
"""

import argparse
import sys
import time

if __name__ == '__main__':
    import psutil

    parser = argparse.ArgumentParser()
    parser.add_argument('--parent-pid', type=int, required=True)
    parser.add_argument('--proc-pid', type=int, required=True)
    args = parser.parse_args()

    process = None
    parent_process = None
    try:
        process = psutil.Process(args.proc_pid)
        parent_process = psutil.Process(args.parent_pid)
    except psutil.NoSuchProcess:
        pass

    if process is None:
        sys.exit()

    if parent_process is not None:
        # Wait for either parent or target process to exit.
        while process.is_running() and parent_process.is_running():
            time.sleep(1)

    try:
        children = process.children(recursive=True)
        children.append(process)
    except psutil.NoSuchProcess:
        sys.exit()

    for pid in children:
        try:
            pid.terminate()
        except psutil.NoSuchProcess:
            pass

    # Wait 30s for the processes to exit gracefully.
    time.sleep(30)

    # SIGKILL if they're still running.
    for pid in children:
        try:
            pid.kill()
        except psutil.NoSuchProcess:
            pass



import copy
import os
from typing import Dict

import yaml

from runhouse.resources.secrets.provider_secrets.provider_secret import ProviderSecret
from runhouse.resources.secrets.utils import _check_file_for_mismatches
from runhouse.utils import create_local_dir


class KubeConfigSecret(ProviderSecret):
    """
    .. note::
        To create a KubeConfigSecret, please use the factory method :func:`provider_secret` with ``provider=="kubernetes"``.
    """

    _PROVIDER = "kubernetes"
    _DEFAULT_CREDENTIALS_PATH = "~/.kube/config"

    @staticmethod
    def from_config(config: dict, dryrun: bool = False, _resolve_children: bool = True):
        # try block if for the case we are trying to load a shared secret.
        return KubeConfigSecret(**config, dryrun=dryrun)

    def _from_path(self, path: str = None):
        path = path or self.path
        if not path:
            return {}

        path = os.path.expanduser(path)
        if os.path.exists(path):
            try:
                with open(path) as f:
                    contents = yaml.safe_load(f)
            except:
                contents = {}
            return contents
        return {}

    def _write_to_file(
        self,
        path: str,
        values: Dict,
        overwrite: bool = False,
        write_config: bool = True,
    ):
        new_secret = copy.deepcopy(self)
        path = path or self.path
        if not _check_file_for_mismatches(
            path, self._from_path(path), values, overwrite
        ):
            full_path = create_local_dir(path)
            with open(full_path, "w") as f:
                yaml.safe_dump(values, f)

            if write_config:
                self._add_to_rh_config(path)

        new_secret._values = None
        new_secret.path = path
        return new_secret



STATIC_CLUSTER_ARGS = {
    "host",
    "ssh_creds",
}

ONDEMAND_COMPUTE_ARGS = {
    "instance_type",
    "num_nodes",
    "provider",
    "use_spot",
    "region",
    "memory",
    "disk_size",
    "vpc_name",
    "num_cpus",
    "accelerators",
    "gpus",
    "sky_kwargs",
    "launcher",
    "autostop_mins",
}

KUBERNETES_CLUSTER_ARGS = {
    "kube_context",
    "kube_namespace",
    "kube_config_path",
}

RH_SERVER_ARGS = {
    "server_port",
    "server_host",
    "ssh_port",
    "open_ports",  # ondemand only
    "server_connection_type",
    "ssl_keyfile",
    "ssl_certfile",
    "domain",
    "image",
}



import pytest

import runhouse as rh

from tests.conftest import init_args


def summer(a: int, b: int):
    print("Running summer function")
    return a + b


def save_and_load_artifacts():
    cpu = rh.ondemand_cluster("^rh-cpu").save()
    loaded_cluster = rh.load(name=cpu.name)
    return loaded_cluster.name


def slow_running_func(a, b):
    import time

    time.sleep(20)
    return a + b


@pytest.fixture(scope="session")
def summer_func(local_launched_ondemand_aws_docker_cluster):
    args = {"name": "summer_func", "fn": summer}
    f = rh.function(**args).to(local_launched_ondemand_aws_docker_cluster)
    init_args[id(f)] = args
    return f


@pytest.fixture(scope="session")
def summer_func_with_auth(ondemand_aws_https_cluster_with_auth):
    return rh.function(summer, name="summer_func").to(
        ondemand_aws_https_cluster_with_auth
    )


@pytest.fixture(scope="session")
def summer_func_shared(shared_cluster):
    return rh.function(summer, name="summer_func").to(shared_cluster)


@pytest.fixture(scope="session")
def func_with_artifacts(local_launched_ondemand_aws_docker_cluster):
    return rh.function(save_and_load_artifacts, name="artifacts_func").to(
        local_launched_ondemand_aws_docker_cluster
    )


@pytest.fixture(scope="session")
def slow_func(local_launched_ondemand_aws_docker_cluster):
    return rh.function(slow_running_func, name="slow_func").to(
        local_launched_ondemand_aws_docker_cluster
    )



import subprocess
import textwrap
from pathlib import Path

from runhouse.constants import DEFAULT_SERVER_PORT
from runhouse.logger import get_logger

logger = get_logger(__name__)
SYSTEMCTL_ERROR = "systemctl: command not found"


class CaddyConfig:
    BASE_CONFIG_PATH = "/etc/caddy/Caddyfile"
    RH_SERVER_PORT = 32300

    # Helpful commands:
    # sudo apt-get install net-tools
    # sudo netstat -tulpn | grep 443

    # For viewing logs:
    # journalctl -u caddy --no-pager | less +G

    # Useful Caddy service commands:
    # sudo systemctl start caddy
    # sudo systemctl stop caddy
    # sudo systemctl status caddy
    # sudo systemctl reload caddy

    # For viewing logs (if running as a service):
    # journalctl -u caddy --no-pager | less +G

    # Useful Caddy background process commands:
    # caddy start
    # caddy stop
    # caddy validate --config /etc/caddy/Caddyfile

    # Checking config settings:
    # caddy adapt --config /etc/caddy/Caddyfile

    def __init__(
        self,
        address: str,
        domain: str = None,
        rh_server_port: int = None,
        ssl_cert_path: str = None,
        ssl_key_path: str = None,
        use_https=False,
        force_reinstall=False,
    ):
        self.use_https = use_https
        self.rh_server_port = rh_server_port or DEFAULT_SERVER_PORT
        self.domain = domain
        self.force_reinstall = force_reinstall

        # To expose the server to the internet, set address to the public IP, otherwise leave it as localhost
        self.address = address or "localhost"

        self.ssl_cert_path = Path(ssl_cert_path).expanduser() if ssl_cert_path else None
        self.ssl_key_path = Path(ssl_key_path).expanduser() if ssl_key_path else None

        if self.use_https and self.domain is None:
            # If using https, need to provide certs
            if self.ssl_cert_path is None:
                raise ValueError(
                    "No SSL cert path provided. Cannot enable HTTPS without a domain or custom certs."
                )

            if not self.ssl_cert_path.exists():
                raise FileNotFoundError(
                    f"Failed to find SSL cert file in path: {self.ssl_cert_path}"
                )

            if self.ssl_key_path is None:
                raise ValueError(
                    "No SSL key path provided. Cannot enable HTTPS without a domain or custom certs."
                )

            if not self.ssl_key_path.exists():
                raise FileNotFoundError(
                    f"Failed to find SSL key file in path: {self.ssl_key_path}"
                )

    @property
    def caddyfile(self):
        """Caddy config file."""
        return Path(self.BASE_CONFIG_PATH).expanduser()

    def configure(self):
        """Configure Caddy to proxy requests to the Fast API HTTP server"""
        if not self._is_configured() or self.force_reinstall:
            logger.debug(f"Configuring Caddy for address: {self.address}")
            self._install()
            self._build_template()
            self._start_caddy()

        # Reload Caddy with the updated config
        logger.debug("Reloading Caddy")
        self.reload()

        if not self._is_configured():
            raise RuntimeError("Failed to configure and start Caddy")

        logger.info("Successfully configured Caddy")

    # -----------------
    # HELPERS
    # -----------------
    def reload(self):
        # https://caddyserver.com/docs/command-line#caddy-stop
        reload_cmd = ["sudo", "systemctl", "reload", "caddy"]
        result = subprocess.run(
            reload_cmd,
            capture_output=True,
            text=True,
        )

        if result.returncode != 0:
            if SYSTEMCTL_ERROR in result.stderr:
                # If running in a docker container or distro without systemctl, reload caddy as a background process
                reload_cmd = f"caddy reload --config {str(self.caddyfile)}"
                try:
                    subprocess.run(reload_cmd, shell=True, check=True, text=True)
                except subprocess.CalledProcessError as e:
                    raise e
            else:
                raise RuntimeError(f"Failed to reload Caddy service: {result.stderr}")

        logger.info("Successfully reloaded Caddy.")

    def _install(self):
        check_cmd = ["sudo", "caddy", "version"]
        result = subprocess.run(
            check_cmd,
            capture_output=True,
            text=True,
        )
        if result.returncode == 0:
            logger.debug("Caddy is already installed, skipping install.")
        else:
            # Install caddy as a service (or background process if we can't use systemctl)
            # https://caddyserver.com/docs/running#using-the-service
            logger.info("Installing Caddy...")

            caddy_version = "2.7.6"
            arch = subprocess.run(
                "dpkg --print-architecture",
                shell=True,
                text=True,
                check=True,
                capture_output=True,
            ).stdout.strip()
            caddy_deb_url = f"https://github.com/caddyserver/caddy/releases/download/v{caddy_version}/caddy_{caddy_version}_linux_{arch}.deb"

            commands = [
                f"sudo wget -q {caddy_deb_url}",
                f'sudo apt install "./caddy_{caddy_version}_linux_{arch}.deb"',
            ]

            for cmd in commands:
                try:
                    subprocess.run(
                        cmd,
                        shell=True,
                        check=True,
                        text=True,
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL,
                    )
                except subprocess.CalledProcessError as e:
                    raise e

        logger.info("Successfully installed Caddy.")

    def _http_template(self):
        return textwrap.dedent(
            f"""
            {{
                default_sni {self.address}
            }}

            http://{self.address} {{
                reverse_proxy localhost:{self.rh_server_port}
            }}
        """
        ).strip()

    def _https_template(self):
        if self.ssl_key_path and self.ssl_cert_path:
            # If custom certs provided use them instead of having Caddy generate them
            # https://caddyserver.com/docs/caddyfile/directives/tls
            logger.info("Using custom certs to enable HTTPs")
            tls_directive = f"tls {self.ssl_cert_path} {self.ssl_key_path}"
            # If domain also provided use it
            address_or_domain = self.domain or self.address
            return textwrap.dedent(
                f"""
                {{
                    default_sni {address_or_domain}
                }}

                https://{address_or_domain} {{
                    {tls_directive}
                    reverse_proxy 127.0.0.1:{self.rh_server_port}
                }}
                """
            ).strip()
        elif self.domain:
            # https://caddyserver.com/docs/automatic-https#hostname-requirements
            logger.info(
                f"Generating certs with Caddy to enable HTTPs using domain: {self.domain}"
            )
            return textwrap.dedent(
                f"""
                https://{self.domain} {{
                    reverse_proxy 127.0.0.1:{self.rh_server_port}
                }}
                """
            ).strip()
        else:
            # Do not support issuing self-signed certs on the cluster
            # Unverified certs should be generated client side and passed to Caddy as custom certs
            raise RuntimeError("No certs or domain provided. Cannot enable HTTPS.")

    def _build_template(self):
        if self.use_https:
            # Update firewall rule for HTTPS
            subprocess.run(
                "sudo ufw allow 443/tcp",
                check=True,
                capture_output=True,
                text=True,
                shell=True,
            )
        else:
            # Update firewall rule for HTTP
            subprocess.run(
                "sudo ufw allow 80/tcp",
                check=True,
                capture_output=True,
                text=True,
                shell=True,
            )

        try:
            template = (
                self._https_template() if self.use_https else self._http_template()
            )

            # Update the Caddyfile with this new template
            subprocess.run(
                f"echo '{template}' | sudo tee {self.caddyfile} > /dev/null",
                shell=True,
                check=True,
                capture_output=True,
                text=True,
            )

        except Exception as e:
            raise e

        logger.info("Successfully built and formatted Caddy template.")

    def _is_configured(self) -> bool:
        logger.debug("Checking Caddy configuration.")
        result = subprocess.run(
            ["sudo", "systemctl", "status", "caddy"],
            capture_output=True,
            text=True,
        )
        if result.returncode != 0:
            if SYSTEMCTL_ERROR in result.stderr:
                # If running in a docker container or distro without systemctl, check whether Caddy has been configured
                run_cmd = f"caddy validate --config {str(self.caddyfile)}"
                try:
                    subprocess.run(
                        run_cmd,
                        shell=True,
                        check=True,
                        text=True,
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL,
                    )
                except subprocess.CalledProcessError:
                    return False
                return True

            else:
                logger.warning(result.stderr)
                return False

        return "active (running)" in result.stdout

    def _start_caddy(self):
        """Run the caddy server as a service or background process. Try to start as a service first, if that
        fails then default to running as a background process."""
        # Add Caddy as a sudoer, otherwise will not be able to install certs on the server
        # Will receive an error that looks like:
        # caddy : user NOT in sudoers ; TTY=unknown ; PWD=/ ; USER=root
        # https://github.com/caddyserver/caddy/issues/4248
        logger.debug("Adding Caddy as trusted app.")
        try:
            subprocess.run(
                "sudo mkdir -p /var/lib/caddy/.local && "
                "sudo chown -R caddy: /var/lib/caddy",
                shell=True,
                check=True,
                text=True,
                stdout=subprocess.DEVNULL,
            )
        except subprocess.CalledProcessError as e:
            raise e

        logger.debug("Starting Caddy.")
        run_cmd = ["sudo", "systemctl", "start", "caddy"]
        result = subprocess.run(
            run_cmd,
            capture_output=True,
            text=True,
        )
        if result.returncode != 0:
            if SYSTEMCTL_ERROR in result.stderr:
                # If running in a docker container or distro without systemctl, we need to start Caddy manually
                # as a background process
                run_cmd = "caddy start"
                try:
                    subprocess.run(
                        run_cmd,
                        shell=True,
                        check=True,
                        text=True,
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL,
                    )
                except subprocess.CalledProcessError as e:
                    raise e
            else:
                raise RuntimeError(f"Failed to run Caddy service: {result.stderr}")

        logger.info("Successfully started Caddy.")



from typing import AsyncIterable, Awaitable, Generator

from runhouse.resources.module import Module
from runhouse.servers.obj_store import RunhouseStopIteration


class FutureModule(Module, Awaitable):
    """A module that represents a future result, but capable of representing a result on a remote system."""

    def __init__(self, future, **kwargs):
        super().__init__(**kwargs)
        self._future = future

    def result(self):
        return self._future.result()

    def remote_await(self):
        return self._future

    def __await__(self):
        # TODO talk about module dunder stuff through this file
        # __await__ needs to return an awaitable, so we wrap it in a coroutine
        async def __await():
            return self.remote_await(run_name=self.name)

        return __await().__await__()

    def set_result(self, __result):
        return self._future.set_result(__result)

    def set_exception(self, __exception):
        return self._future.set_exception(__exception)

    def done(self):
        return self._future.done()

    def cancelled(self):
        return self._future.cancelled()

    def add_done_callback(self, __fn, *, context=None):
        return self._future.add_done_callback(__fn, context=context)

    def remove_done_callback(self, __fn):
        return self._future.remove_done_callback(__fn)

    def cancel(self, msg=None):
        return self._future.cancel(msg=msg)

    def exception(self):
        return self._future.exception()

    def __getstate__(self):
        state = super().__getstate__()
        state["_future"] = None
        return state


class GeneratorModule(Module, Generator):
    """A module that represents a future result, but capable of representing a result on a remote system."""

    def __init__(self, future, **kwargs):
        super().__init__(**kwargs)
        self._future = future

    def __iter__(self):
        # Note that we can't return self._future.__iter__() here because that would be a generator,
        # which we can't pickle. Instead, we use __next__-style generation.
        return self

    def remote_next(self):
        try:
            return self._future.__next__()
        except StopIteration:
            raise RunhouseStopIteration()

    def __next__(self):
        return self.remote_next()

    def send(self, __value):
        return self._future.send(__value)

    def throw(self, __typ, __val=None, __tb=None):
        return self._future.throw(__typ, __val, __tb)

    def __getstate__(self):
        state = super().__getstate__()
        state["_future"] = None
        return state


class AsyncGeneratorModule(Module, AsyncIterable):
    """A module that represents a future result, but capable of representing a result on a remote system."""

    def __init__(self, future, **kwargs):
        super().__init__(**kwargs)
        self._future = future

    def __aiter__(self):
        return self

    def remote_anext(self):
        return self._future.__anext__()

    async def __anext__(self):
        return self.remote_anext(run_name=self.name)

    async def remote_await(self):
        return await self._future

    def __await__(self):
        return self.remote_await()

    def send(self, __value):
        return self._future.send(__value)

    def throw(self, __typ, __val=None, __tb=None):
        return self._future.throw(__typ, __val, __tb)

    def __getstate__(self):
        state = super().__getstate__()
        state["_future"] = None
        return state



import pytest

from runhouse.globals import rns_client
from runhouse.servers.obj_store import ObjStoreError

from tests.utils import friend_account, get_ray_servlet_and_obj_store


def list_compare(list1, list2):
    return sorted(list1) == sorted(list2)


OBJ_STORE_NAME = "test_obj_store"


@pytest.mark.servertest
@pytest.mark.parametrize("obj_store", [OBJ_STORE_NAME], indirect=True)
class TestObjStore:
    """Start object store in a local base servlet"""

    @pytest.mark.level("unit")
    @pytest.mark.parametrize("key", ["k1", 123])
    @pytest.mark.parametrize(
        "value",
        ["test", 1, 1.0, True, False, None, b"\x00\x01\x02", [1, 2, 3], ["1", 1, 1.5]],
    )
    def test_put_get_delete_string(self, obj_store, key, value):
        assert obj_store.keys() == []

        key = "k1"
        obj_store.put(key, value)
        assert obj_store.keys() == [key]

        res = obj_store.get(key)
        assert res == value
        assert obj_store.get_servlet_name_for_key(key) == obj_store.servlet_name

        obj_store.put(key, "overwrite_value")
        assert obj_store.keys() == [key]

        res = obj_store.get(key)
        assert res == "overwrite_value"
        assert obj_store.get_servlet_name_for_key(key) == obj_store.servlet_name

        obj_store.delete(key)
        assert obj_store.keys() == []
        assert obj_store.get(key, default=None) is None
        assert obj_store.get_servlet_name_for_key(key) is None

    @pytest.mark.level("unit")
    @pytest.mark.parametrize("default_value", [1, None, "asdf"])
    def test_get_default_behavior(self, obj_store, default_value):
        assert obj_store.keys() == []

        assert obj_store.get("a", default=default_value) == default_value
        assert obj_store.get("b", default=default_value) == default_value

    @pytest.mark.level("unit")
    def test_get_default_key_error(self, obj_store):
        assert obj_store.keys() == []

        try:
            obj_store.get("c", default=KeyError)
        except KeyError:
            return

        assert False, "Should have raised KeyError"

    @pytest.mark.level("unit")
    def test_pop(self, obj_store):
        assert obj_store.keys() == []

        key = "new_key"
        obj_store.put(key, "v1")
        value = obj_store.pop(key)
        assert value == "v1"

        res = obj_store.get(key, default=None)
        assert res is None

    @pytest.mark.level("unit")
    def test_get_list(self, obj_store):
        assert obj_store.keys() == []

        keys = ["k1", "k2", "k3"]
        vals = ["v1", "v2", "v3"]

        for k, v in zip(keys, vals):
            obj_store.put(k, v)

        res = obj_store.get_list(keys)
        assert res == ["v1", "v2", "v3"]

    @pytest.mark.level("unit")
    def test_rename(self, obj_store):
        assert obj_store.keys() == []

        key = "k1"
        new_key = "k2"
        val = "v1"

        obj_store.put(key, val)
        obj_store.rename(key, new_key)
        res = obj_store.get(new_key)
        assert res == val
        assert obj_store.get(key, default=None) is None
        assert obj_store.keys() == [new_key]
        assert obj_store.get_servlet_name_for_key(new_key) == obj_store.servlet_name
        assert obj_store.get_servlet_name_for_key(key) is None

    @pytest.mark.level("unit")
    def test_clear(self, obj_store):
        assert obj_store.keys() == []

        keys = ["k1", "k2", "k3"]
        vals = ["v1", "v2", "v3"]

        for k, v in zip(keys, vals):
            obj_store.put(k, v)

        assert obj_store.keys() == keys

        obj_store.clear()
        assert obj_store.keys() == []

    @pytest.mark.level("unit")
    def test_many_servlets(self, obj_store):
        assert obj_store.keys() == []

        _, obj_store_2 = get_ray_servlet_and_obj_store("other")
        assert obj_store_2.keys() == []

        obj_store.put("k1", "v1")
        obj_store_2.put("k2", "v2")
        obj_store_2.put("k3", "v3")

        assert list_compare(obj_store.keys(), ["k1", "k2", "k3"])
        assert list_compare(obj_store_2.keys(), ["k1", "k2", "k3"])

        assert obj_store.get("k1") == "v1"
        assert obj_store.get("k2") == "v2"
        assert obj_store.get("k3") == "v3"
        assert obj_store_2.get("k1") == "v1"
        assert obj_store_2.get("k2") == "v2"
        assert obj_store_2.get("k3") == "v3"

        assert obj_store.get_servlet_name_for_key("k1") == obj_store.servlet_name
        assert obj_store.get_servlet_name_for_key("k2") == obj_store_2.servlet_name
        assert obj_store.get_servlet_name_for_key("k3") == obj_store_2.servlet_name
        assert obj_store_2.get_servlet_name_for_key("k1") == obj_store.servlet_name
        assert obj_store_2.get_servlet_name_for_key("k2") == obj_store_2.servlet_name
        assert obj_store_2.get_servlet_name_for_key("k3") == obj_store_2.servlet_name

        # Technically, "k1" is only present on the base servlet,
        # and "k2" and "k3" are only present on the other servlet
        # These methods are static, we can run them from either store
        assert obj_store.keys_for_servlet_name(obj_store.servlet_name) == ["k1"]
        assert obj_store.get_from_servlet_name(obj_store.servlet_name, "k1") == "v1"
        assert obj_store.get_from_servlet_name(obj_store.servlet_name, "k2") is None
        assert obj_store.get_from_servlet_name(obj_store.servlet_name, "k3") is None

        assert obj_store.keys_for_servlet_name(obj_store_2.servlet_name) == [
            "k2",
            "k3",
        ]
        assert obj_store.get_from_servlet_name(obj_store_2.servlet_name, "k1") is None
        assert obj_store.get_from_servlet_name(obj_store_2.servlet_name, "k2") == "v2"
        assert obj_store.get_from_servlet_name(obj_store_2.servlet_name, "k3") == "v3"

        # Overwriting "k2" from obj_store instead of obj_store_2
        obj_store.put("k2", "changed")
        assert list_compare(obj_store.keys(), ["k1", "k3", "k2"])
        assert obj_store.get("k2") == "changed"
        assert obj_store_2.get("k2") == "changed"
        assert obj_store.get_servlet_name_for_key("k2") == obj_store.servlet_name
        assert obj_store_2.get_servlet_name_for_key("k2") == obj_store.servlet_name

        assert list_compare(
            obj_store.keys_for_servlet_name(obj_store.servlet_name),
            [
                "k1",
                "k2",
            ],
        )
        assert obj_store.get_from_servlet_name(obj_store.servlet_name, "k1") == "v1"
        assert (
            obj_store.get_from_servlet_name(obj_store.servlet_name, "k2") == "changed"
        )
        assert obj_store.get_from_servlet_name(obj_store.servlet_name, "k3") is None

        assert obj_store.keys_for_servlet_name(obj_store_2.servlet_name) == ["k3"]
        assert obj_store.get_from_servlet_name(obj_store_2.servlet_name, "k1") is None
        assert obj_store.get_from_servlet_name(obj_store_2.servlet_name, "k2") is None
        assert obj_store.get_from_servlet_name(obj_store_2.servlet_name, "k3") == "v3"

        # Renaming "k2" to "key_changed" from obj_store_2
        # Even though "k2" is technically on base object store.
        obj_store_2.rename("k2", "key_changed")
        assert list_compare(obj_store.keys(), ["k1", "k3", "key_changed"])
        assert obj_store.get("key_changed") == "changed"
        assert obj_store_2.get("key_changed") == "changed"
        assert (
            obj_store.get_servlet_name_for_key("key_changed") == obj_store.servlet_name
        )
        assert (
            obj_store_2.get_servlet_name_for_key("key_changed")
            == obj_store.servlet_name
        )

        assert list_compare(
            obj_store.keys_for_servlet_name(obj_store.servlet_name),
            [
                "k1",
                "key_changed",
            ],
        )
        assert obj_store.get_from_servlet_name(obj_store.servlet_name, "k1") == "v1"
        assert (
            obj_store.get_from_servlet_name(obj_store.servlet_name, "key_changed")
            == "changed"
        )
        assert obj_store.get_from_servlet_name(obj_store.servlet_name, "k3") is None

        assert obj_store.keys_for_servlet_name(obj_store_2.servlet_name) == ["k3"]
        assert obj_store.get_from_servlet_name(obj_store_2.servlet_name, "k1") is None
        assert (
            obj_store.get_from_servlet_name(obj_store_2.servlet_name, "key_changed")
            is None
        )
        assert obj_store.get_from_servlet_name(obj_store_2.servlet_name, "k3") == "v3"

        # Renaming "key_changed" to "k3"
        obj_store.rename("key_changed", "k3")
        assert list_compare(obj_store.keys(), ["k1", "k3"])
        assert obj_store.get("k3") == "changed"
        assert obj_store_2.get("k3") == "changed"
        assert obj_store.get_servlet_name_for_key("k3") == obj_store.servlet_name
        assert obj_store_2.get_servlet_name_for_key("k3") == obj_store.servlet_name

        assert obj_store.keys_for_servlet_name(obj_store.servlet_name) == [
            "k1",
            "k3",
        ]
        assert obj_store.get_from_servlet_name(obj_store.servlet_name, "k1") == "v1"
        assert (
            obj_store.get_from_servlet_name(obj_store.servlet_name, "k3") == "changed"
        )
        assert (
            obj_store.get_from_servlet_name(obj_store.servlet_name, "key_changed")
            is None
        )

        assert obj_store.keys_for_servlet_name(obj_store_2.servlet_name) == []
        assert obj_store.get_from_servlet_name(obj_store_2.servlet_name, "k1") is None
        assert obj_store.get_from_servlet_name(obj_store_2.servlet_name, "k3") is None
        assert (
            obj_store.get_from_servlet_name(obj_store_2.servlet_name, "key_changed")
            is None
        )

        # Popping "k3" from obj_store_2
        res = obj_store_2.pop("k3")
        assert res == "changed"
        assert obj_store.keys() == ["k1"]
        assert obj_store.get("k3") is None
        assert obj_store_2.get("k3") is None
        assert obj_store.get_servlet_name_for_key("k3") is None
        assert obj_store_2.get_servlet_name_for_key("k3") is None

        assert obj_store.keys_for_servlet_name(obj_store.servlet_name) == ["k1"]
        assert obj_store.get_from_servlet_name(obj_store.servlet_name, "k1") == "v1"
        assert obj_store.keys_for_servlet_name(obj_store_2.servlet_name) == []

        # Testing of maintaining processes
        _, obj_store_3 = get_ray_servlet_and_obj_store("third")
        assert obj_store_3.keys() == ["k1"]
        obj_store_3.put("k2", "v2")
        obj_store_3.put("k3", "v3")
        assert list_compare(obj_store_3.keys(), ["k1", "k2", "k3"])
        assert list_compare(
            obj_store_3.keys_for_servlet_name(obj_store_3.servlet_name),
            ["k2", "k3"],
        )
        obj_store_2.clear()
        assert obj_store_3.keys() == []

    @pytest.mark.level("unit")
    def test_delete_servlet(self, obj_store):
        _, obj_store_2 = get_ray_servlet_and_obj_store("obj_store_2")

        assert obj_store.keys() == []
        assert obj_store_2.keys() == []

        obj_store.put("k1", "v1")
        obj_store_2.put("k2", "v2")
        obj_store_2.put("k3", "v3")

        servlet_to_delete = obj_store_2.servlet_name
        obj_store_2_keys = obj_store_2.keys_for_servlet_name(servlet_to_delete)

        assert servlet_to_delete in obj_store.get_all_initialized_servlet_args()
        for key in obj_store_2_keys:
            assert obj_store.get(key)

        obj_store.delete_servlet_contents(servlet_to_delete)

        # check obj_store_2 servlet and nested keys are deleted but obj_store_1 unaffected
        assert servlet_to_delete not in obj_store.get_all_initialized_servlet_args()
        for key in obj_store_2_keys:
            assert not obj_store.get(key)
        assert obj_store.servlet_name in obj_store.get_all_initialized_servlet_args()
        assert obj_store.get("k1")

        # check that corresponding Ray actor is killed
        with pytest.raises(ObjStoreError):
            obj_store.get_servlet(name=servlet_to_delete, raise_ex_if_not_found=True)


@pytest.mark.servertest
@pytest.mark.parametrize("obj_store", [OBJ_STORE_NAME], indirect=True)
class TestAuthCacheObjStore:
    """Start object store in a local auth cache servlet"""

    @pytest.mark.level("unit")
    def test_save_resources_to_obj_store_cache(self, obj_store):
        with friend_account() as test_account_dict:
            token = test_account_dict["token"]
            username = test_account_dict["username"]
            default_folder = test_account_dict["default_folder"]

            # Use cluster token when validating access
            cluster_token = rns_client.cluster_token(
                resource_address=default_folder, username=username, den_token=token
            )

            # Add test account resources to the local cache
            resource_uri = f"/{test_account_dict['username']}/summer"
            access_level = obj_store.resource_access_level(cluster_token, resource_uri)

            assert access_level == "write"

    @pytest.mark.level("unit")
    def test_no_resource_access_for_invalid_token(self, obj_store):
        with friend_account() as test_account_dict:
            cluster_token = "abc"
            resource_uri = f"/{test_account_dict['username']}/summer"
            access_level = obj_store.resource_access_level(cluster_token, resource_uri)
            assert access_level is None



from runhouse.constants import DEFAULT_DASK_PORT
from runhouse.resources.distributed.supervisor import Supervisor

from runhouse.resources.module import Module


class DaskDistributed(Supervisor):
    def __init__(
        self,
        name,
        module: Module = None,
        port: int = DEFAULT_DASK_PORT,
        client_timeout="3s",
        **kwargs
    ):
        super().__init__(name=name, **kwargs)
        self._module = module
        self._dask_port = port
        self._dask_client = None
        self._client_timeout = client_timeout

    def _compute_signature(self, rich=False):
        return self.local._module.signature(rich=rich)

    def forward(self, item, *args, **kwargs):
        if not self._dask_client:
            self._dask_client = self.system.connect_dask(
                port=self._dask_port, client_timeout=self._client_timeout
            )
        method = getattr(self._module, item)
        return method(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        return self.call(*args, **kwargs)

    def __getstate__(self):
        state = super().__getstate__()
        # Dask client can't be serialized
        state["_dask_client"] = None
        return state



import copy
import json

from unittest.mock import AsyncMock, patch

import pytest
import runhouse as rh
from runhouse.constants import DEFAULT_PROCESS_NAME, DEFAULT_SERVER_PORT

from runhouse.globals import rns_client

from runhouse.servers.http import HTTPClient
from runhouse.servers.http.http_utils import (
    DeleteObjectParams,
    PutObjectParams,
    serialize_data,
)


@pytest.mark.servertest
class TestHTTPClient:
    @pytest.fixture(autouse=True)
    def init_fixtures(self):
        from tests.fixtures.secret_fixtures import provider_secret_values

        args = dict(
            name="local-cluster",
            host="localhost",
            server_host="0.0.0.0",
            ssh_creds=rh.provider_secret(
                provider="ssh", values=provider_secret_values["ssh"]
            ),
        )
        self.local_cluster = rh.cluster(**args)
        self.client = HTTPClient(
            "localhost",
            DEFAULT_SERVER_PORT,
            resource_address=self.local_cluster.rns_address,
        )

    @pytest.mark.level("unit")
    def test_check_server(self, mocker):
        mock_response = mocker.Mock()
        mock_response.status_code = 200
        rh_version_resp = {"rh_version": rh.__version__}
        mock_response.json.return_value = rh_version_resp
        mocked_get = mocker.patch("requests.Session.get", return_value=mock_response)

        self.client.check_server()
        expected_verify = self.client.verify

        mocked_get.assert_called_once_with(
            f"http://localhost:{DEFAULT_SERVER_PORT}/check",
            timeout=HTTPClient.CHECK_TIMEOUT_SEC,
            verify=expected_verify,
        )

    @pytest.mark.level("unit")
    def test_get_certificate(self, mocker):

        mock_request = mocker.patch("runhouse.servers.http.HTTPClient.request")
        mock_request.return_value = b"certificate_content"

        # Set up the mocker for 'mkdir' method
        mock_mkdir = mocker.patch("pathlib.Path.mkdir")

        # Set up the mocker for 'open' method
        mock_file_open = mocker.patch(
            "builtins.open",
            new_callable=mocker.mock_open,
            read_data="certificate_content",
        )

        self.client.cert_path = "/fake/path/cert.pem"

        self.client.get_certificate()

        mock_request.assert_called_once_with("cert", req_type="get", headers={})

        mock_mkdir.assert_called_once()

        # Check that open was called correctly
        mock_file_open.assert_called_once_with("/fake/path/cert.pem", "wb")

        # Check that the correct content was written to the file
        mock_file_open().write.assert_called_once_with(b"certificate_content")

    @pytest.mark.level("unit")
    @patch("runhouse.globals.rns_client.request_headers")
    def test_use_cert_verification(self, mock_request_headers, mocker):
        # Mock the request_headers to avoid actual HTTP requests in the test for loading the cluster token
        mock_request_headers.return_value = {"Authorization": "Bearer mock_token"}

        # Mock a certificate where the issuer is different from the subject
        mock_cert = mocker.MagicMock()
        mock_cert.issuer = "issuer"
        mock_cert.subject = "subject"

        mock_load_cert = mocker.patch(
            "builtins.open", mocker.mock_open(read_data="cert_data")
        )
        mocker.patch(
            "cryptography.x509.load_pem_x509_certificate", return_value=mock_cert
        )

        mock_load_cert.return_value = mock_cert

        # Test with HTTPS enabled and a valid cert path which is not self-signed
        client = HTTPClient(
            "localhost",
            DEFAULT_SERVER_PORT,
            resource_address=self.local_cluster.rns_address,
            use_https=True,
            cert_path="/valid/path",
        )
        assert client.verify is True

        # Mock a self-signed cert where the issuer is the same as the subject
        mock_cert.issuer = "self-signed"
        mock_cert.subject = "self-signed"
        mock_load_cert.return_value = mock_cert

        # If providing a valid self-signed cert, "verify" should be the path to the cert
        mocker.patch("pathlib.Path.exists", return_value=True)
        client = HTTPClient(
            "localhost",
            DEFAULT_SERVER_PORT,
            resource_address=self.local_cluster.rns_address,
            use_https=True,
            cert_path="/self-signed/path",
        )
        assert client.verify == "/self-signed/path"

        # If providing an invalid cert path we still default to verify=True since https is enabled
        mocker.patch("pathlib.Path.exists", return_value=False)
        client = HTTPClient(
            "localhost",
            DEFAULT_SERVER_PORT,
            resource_address=self.local_cluster.rns_address,
            use_https=True,
            cert_path="/invalid/path",
        )
        assert client.verify is True

    @pytest.mark.level("unit")
    def test_call_module_method(self, mocker):
        expected_headers = rns_client.request_headers(
            resource_address=self.local_cluster.rns_address
        )
        # Mock the response to iter_lines to return our simulated server response
        mock_response = mocker.Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "output_type": "result_serialized",
            "data": serialize_data("final_result", "pickle"),
            "serialization": "pickle",
        }
        mock_post = mocker.patch("requests.Session.post", return_value=mock_response)

        # Mock response to the logs function separately
        mock_logs_aiter = AsyncMock()
        mock_logs_aiter.__aiter__.return_value = iter(
            [
                json.dumps({"output_type": "stdout", "data": "Log message"}),
            ]
        )

        mock_logs_iter = mocker.Mock()
        mock_logs_iter.aiter_lines.return_value = mock_logs_aiter
        mock_logs_iter.status_code = 200

        mock_stream_get = AsyncMock()
        mock_stream_get.__aenter__.return_value = mock_logs_iter

        _ = mocker.patch("httpx.AsyncClient.stream", return_value=mock_stream_get)

        # Call the method under test
        method_name = "install"
        module_name = DEFAULT_PROCESS_NAME

        # Need to specify the run_name to avoid generating a unique one that contains the timestamp
        result = self.client.call(
            module_name,
            method_name,
            run_name="test_run_name",
        )

        assert result == "final_result"

        # Assert that the post request was called correctly
        expected_url = self.client._formatted_url(f"{module_name}/{method_name}")
        expected_json_data = {
            "data": None,
            "serialization": "pickle",
            "run_name": "test_run_name",
            "stream_logs": True,
            "save": False,
            "remote": False,
        }

        expected_verify = self.client.verify
        mock_post.assert_called_once_with(
            expected_url,
            json=expected_json_data,
            headers=expected_headers,
            auth=None,
            verify=expected_verify,
        )

    @pytest.mark.level("unit")
    def test_call_module_method_with_args_kwargs(self, mocker):
        expected_headers = rns_client.request_headers(self.local_cluster.rns_address)
        # Mock the response to iter_lines to return our simulated server response
        mock_response = mocker.Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "output_type": "success",
            "data": None,
            "serialization": None,
        }
        mock_post = mocker.patch("requests.Session.post", return_value=mock_response)

        # Mock response to the logs function separately
        mock_logs_aiter = AsyncMock()
        mock_logs_aiter.__aiter__.return_value = iter(
            [
                json.dumps({"output_type": "stdout", "data": "Log message"}),
            ]
        )

        mock_logs_iter = mocker.Mock()
        mock_logs_iter.aiter_lines.return_value = mock_logs_aiter
        mock_logs_iter.status_code = 200

        mock_stream_get = AsyncMock()
        mock_stream_get.__aenter__.return_value = mock_logs_iter

        _ = mocker.patch("httpx.AsyncClient.stream", return_value=mock_stream_get)

        data = {"args": [1, 2], "kwargs": {"a": 3, "b": 4}}
        module_name = "module"
        method_name = "install"

        # Need to specify the run_name to avoid generating a unique one that contains the timestamp
        self.client.call(
            module_name,
            method_name,
            data=data,
            run_name="test_run_name",
        )

        # Assert that the post request was called with the correct data
        expected_json_data = {
            "data": serialize_data(data, "pickle"),
            "serialization": "pickle",
            "run_name": "test_run_name",
            "stream_logs": True,
            "save": False,
            "remote": False,
        }
        expected_url = f"http://localhost:32300/{module_name}/{method_name}"
        expected_verify = self.client.verify

        mock_post.assert_called_with(
            expected_url,
            json=expected_json_data,
            headers=expected_headers,
            auth=None,
            verify=expected_verify,
        )

    @pytest.mark.level("unit")
    def test_call_module_method_error_handling(self, mocker, local_cluster):
        mock_response = mocker.Mock()
        mock_response.status_code = 500
        mock_response.content = b"Internal Server Error"
        mocker.patch("requests.Session.post", return_value=mock_response)

        with pytest.raises(ValueError):
            self.client.call("module", "method")

    @pytest.mark.level("unit")
    def test_call_module_method_config(self, mocker, local_cluster):
        request_headers = rns_client.request_headers(local_cluster.rns_address)

        test_data = self.local_cluster.config()
        mock_response = mocker.Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "output_type": "config",
            "data": copy.copy(test_data),
        }
        _ = mocker.patch("requests.Session.post", return_value=mock_response)

        cluster = self.client.call(
            DEFAULT_PROCESS_NAME,
            "install",
            headers=request_headers,
            stream_logs=False,
        )
        assert cluster.config() == test_data

    @pytest.mark.level("unit")
    def test_put_object(self, mocker):

        mock_request = mocker.patch("runhouse.servers.http.HTTPClient.request_json")

        key = "my_list"
        value = list(range(5, 50, 2)) + ["a string"]
        expected_data = serialize_data(value, "pickle")

        self.client.put_object(key, value)

        mock_request.assert_called_once_with(
            "object",
            req_type="post",
            json_dict=mocker.ANY,
            err_str=f"Error putting object {key}",
        )

        actual_data = PutObjectParams(**mock_request.call_args[1]["json_dict"])
        assert actual_data.key == key
        assert actual_data.serialized_data == expected_data
        assert actual_data.serialization == "pickle"

    @pytest.mark.level("unit")
    def test_get_keys(self, mocker):
        mock_request = mocker.patch("runhouse.servers.http.HTTPClient.request")

        self.client.keys()
        mock_request.assert_called_with("keys", req_type="get")

        mock_request.reset_mock()

        test_process = "test_process"
        self.client.keys(process=test_process)
        mock_request.assert_called_with(
            f"keys/?process_name={test_process}", req_type="get"
        )

    @pytest.mark.level("unit")
    def test_delete(self, mocker):

        mock_request = mocker.patch("runhouse.servers.http.HTTPClient.request_json")

        keys = ["key1", "key2"]

        self.client.delete(keys=keys)

        mock_request.assert_called_once_with(
            "delete_object",
            req_type="post",
            json_dict=mocker.ANY,
            err_str=f"Error deleting keys {keys}",
        )

        actual_data = DeleteObjectParams(**mock_request.call_args[1]["json_dict"])
        assert actual_data.keys == keys
